import tempfile
from collections.abc import Callable

import polars as pl
import pytest
from httpx import Response
from polars.testing import assert_frame_equal
from pyarrow import Table
from respx import MockRouter
from sqlalchemy import Engine

from matchbox.client.dags import DAG
from matchbox.client.eval import get_samples
from matchbox.client.models.linkers import DeterministicLinker
from matchbox.client.results import ResolvedMatches
from matchbox.common.arrow import (
    SCHEMA_EVAL_SAMPLES,
    SCHEMA_QUERY_WITH_LEAVES,
    table_to_buffer,
)
from matchbox.common.dtos import Collection, Resolution, ResolutionType, Run, User
from matchbox.common.exceptions import MatchboxSourceTableError
from matchbox.common.factories.dags import TestkitDAG
from matchbox.common.factories.sources import source_from_tuple


def test_get_samples_local(sqlite_in_memory_warehouse: Engine) -> None:
    """We can sample from a parquet dump."""
    dag = DAG("companies")
    foo = dag.source(
        **source_from_tuple(
            name="foo",
            engine=sqlite_in_memory_warehouse,
            data_keys=["1", "2", "2b", "3"],
            data_tuple=(
                {"field_a": 10},
                {"field_a": 20},
                {"field_a": 20},
                {"field_a": 30},
            ),
        )
        .write_to_location()
        .into_dag()
    )
    bar = dag.source(
        **source_from_tuple(
            name="bar",
            engine=sqlite_in_memory_warehouse,
            data_keys=["a", "b", "c", "d"],
            data_tuple=(
                {"field_a": "1x", "field_b": "1y"},
                {"field_a": "2x", "field_b": "2y"},
                {"field_a": "3x", "field_b": "3y"},
                {"field_a": "4x", "field_b": "4y"},
            ),
        )
        .write_to_location()
        .into_dag()
    )

    # Both foo and bar have a record that's not linked
    # Foo has two keys for one leaf ID
    # Foo and bar have links across; bar also has link within
    foo_query_data = pl.DataFrame(
        [
            {"id": 14, "leaf_id": 1, "key": "1"},
            {"id": 2, "leaf_id": 2, "key": "2"},
            {"id": 2, "leaf_id": 2, "key": "2b"},
            {"id": 356, "leaf_id": 3, "key": "3"},
        ],
        schema=pl.Schema(SCHEMA_QUERY_WITH_LEAVES),
    )

    bar_query_data = pl.DataFrame(
        [
            {"id": 14, "leaf_id": 4, "key": "a"},
            {"id": 356, "leaf_id": 5, "key": "b"},
            {"id": 356, "leaf_id": 6, "key": "c"},
            {"id": 7, "leaf_id": 7, "key": "d"},
        ],
        schema=pl.Schema(SCHEMA_QUERY_WITH_LEAVES),
    )

    rm = ResolvedMatches(
        sources=[foo, bar], query_results=[foo_query_data, bar_query_data]
    )

    # Create a temporary file with .pq suffix
    with tempfile.NamedTemporaryFile(suffix=".pq") as tmp_file:
        # Write the parquet data to the temporary file
        rm.as_dump().write_parquet(tmp_file.name)

        # Use the temporary file in get_samples
        samples = get_samples(
            n=2, dag=dag, user_name="alice", sample_file=tmp_file.name
        )
        assert len(samples) == 2
        possible_clusters = set(foo_query_data["id"]) | set(bar_query_data["id"])
        assert set(samples.keys()) <= possible_clusters


def test_get_samples_remote(
    matchbox_api: MockRouter,
    sqla_sqlite_warehouse: Engine,
    sqlite_in_memory_warehouse: Engine,
    env_setter: Callable[[str, str], None],
) -> None:
    """We can sample from a resolution on the server."""
    # Make dummmy data
    user = User(user_name="alice", email="alice@example.com")

    # Foo has two identical rows
    foo_testkit = source_from_tuple(
        data_tuple=({"col": 1}, {"col": 1}, {"col": 2}, {"col": 3}, {"col": 4}),
        data_keys=["1", "1bis", "2", "3", "4"],
        name="foo",
        location_name="db",
        engine=sqla_sqlite_warehouse,
    ).write_to_location()

    bar_testkit = source_from_tuple(
        data_tuple=({"col": 1}, {"col": 2}, {"col": 3}, {"col": 4}),
        data_keys=["a", "b", "c", "d"],
        name="bar",
        location_name="db",
        engine=sqla_sqlite_warehouse,
    ).write_to_location()

    # This will be excluded as the location name differs
    baz_testkit = source_from_tuple(
        data_tuple=({"col": 1},),
        data_keys=["x"],
        name="baz",
        location_name="db_other",
        engine=sqla_sqlite_warehouse,
    ).write_to_location()

    dag = TestkitDAG().dag

    foo = dag.source(**foo_testkit.into_dag())
    bar = dag.source(**bar_testkit.into_dag())
    baz = dag.source(**baz_testkit.into_dag())
    foo_bar = foo.query().linker(
        bar.query(),
        name="linker1",
        model_class=DeterministicLinker,
        model_settings={"comparisons": "l.key=r.key"},
    )
    foo_bar_baz = foo_bar.query(foo, bar).linker(
        baz.query(),
        name="linker2",
        model_class=DeterministicLinker,
        model_settings={"comparisons": "l.key=r.key"},
    )

    # Mock the collection and run endpoint that load_pending() calls
    collection_data = Collection(runs=[dag.run])
    run_data = Run(
        run_id=dag.run,
        mutable=True,
        resolutions={
            "foo": foo_testkit.fake_run().source.to_resolution(),
            "bar": bar_testkit.fake_run().source.to_resolution(),
            "baz": baz_testkit.fake_run().source.to_resolution(),
            "linker1": Resolution(
                fingerprint=b"mock",
                truth=1,
                resolution_type=ResolutionType.MODEL,
                config=foo_bar.config,
            ),
            "linker2": Resolution(
                fingerprint=b"mock2",
                truth=1,
                resolution_type=ResolutionType.MODEL,
                config=foo_bar_baz.config,
            ),
        },
    )

    # Mock API endpoints for load_pending
    matchbox_api.get(f"/collections/{dag.name}").mock(
        return_value=Response(200, content=collection_data.model_dump_json())
    )
    matchbox_api.get(f"/collections/{dag.name}/runs/{dag.run}").mock(
        return_value=Response(200, content=run_data.model_dump_json())
    )

    # Mock samples
    samples = Table.from_pylist(
        [
            # Source foo - with two keys for one leaf
            {"root": 10, "leaf": 1, "key": "1", "source": "foo"},
            {"root": 10, "leaf": 1, "key": "1bis", "source": "foo"},
            {"root": 10, "leaf": 2, "key": "2", "source": "foo"},
            {"root": 11, "leaf": 3, "key": "3", "source": "foo"},
            {"root": 11, "leaf": 4, "key": "4", "source": "foo"},
            # Source bar
            {"root": 10, "leaf": 5, "key": "a", "source": "bar"},
            {"root": 10, "leaf": 6, "key": "b", "source": "bar"},
            {"root": 11, "leaf": 7, "key": "c", "source": "bar"},
            {"root": 11, "leaf": 8, "key": "d", "source": "bar"},
            # Source baz
            {"root": 10, "leaf": 1, "key": "x", "source": "baz"},
        ],
        schema=SCHEMA_EVAL_SAMPLES,
    )
    # There will be nulls in case of a schema mismatch
    assert len(samples.drop_null()) == len(samples)

    matchbox_api.get("/eval/samples").mock(
        return_value=Response(200, content=table_to_buffer(samples).read())
    )

    # Create a fresh DAG and load it with warehouse location
    # (can't reuse the existing dag as it already has sources added)

    loaded_dag: DAG = (
        DAG(name=str(dag.name)).load_pending().set_client(sqla_sqlite_warehouse)
    )

    # Check results - test with samples that include all three sources
    # All three sources (foo, bar, baz) are in loaded_dag with the warehouse location
    samples_all = get_samples(
        n=10,
        resolution=dag.final_step.resolution_path.name,
        user_name=user.user_name,
        dag=loaded_dag,
    )

    assert sorted(samples_all.keys()) == [10, 11]

    # Now test with samples that only include foo and bar (not baz)
    samples_no_baz = Table.from_pylist(
        [
            # Source foo - with two keys for one leaf
            {"root": 10, "leaf": 1, "key": "1", "source": "foo"},
            {"root": 10, "leaf": 1, "key": "1bis", "source": "foo"},
            {"root": 10, "leaf": 2, "key": "2", "source": "foo"},
            {"root": 11, "leaf": 3, "key": "3", "source": "foo"},
            {"root": 11, "leaf": 4, "key": "4", "source": "foo"},
            # Source bar
            {"root": 10, "leaf": 5, "key": "a", "source": "bar"},
            {"root": 10, "leaf": 6, "key": "b", "source": "bar"},
            {"root": 11, "leaf": 7, "key": "c", "source": "bar"},
            {"root": 11, "leaf": 8, "key": "d", "source": "bar"},
        ],
        schema=SCHEMA_EVAL_SAMPLES,
    )
    matchbox_api.get("/eval/samples").mock(
        return_value=Response(200, content=table_to_buffer(samples_no_baz).read())
    )

    samples = get_samples(
        n=10,
        resolution=dag.final_step.resolution_path.name,
        user_name=user.user_name,
        dag=loaded_dag,
    )

    assert sorted(samples.keys()) == [10, 11]

    expected_sample_10 = pl.DataFrame(
        {
            "leaf": [1, 1, 2, 5, 6],
            "foo_col": [1, 1, 2, None, None],
            "bar_col": [None, None, None, 1, 2],
        }
    )

    expected_sample_11 = pl.DataFrame(
        {
            "leaf": [3, 4, 7, 8],
            "foo_col": [3, 4, None, None],
            "bar_col": [None, None, 3, 4],
        }
    )

    # EvaluationItems.records contains the data with qualified column names
    assert_frame_equal(
        samples[10].records,
        expected_sample_10,
        check_column_order=False,
        check_row_order=False,
        check_dtypes=False,
    )
    assert_frame_equal(
        samples[11].records,
        expected_sample_11,
        check_column_order=False,
        check_row_order=False,
        check_dtypes=False,
    )

    # What happens if no samples are available?
    matchbox_api.get("/eval/samples").mock(
        return_value=Response(
            200,
            content=table_to_buffer(
                Table.from_pylist([], schema=SCHEMA_EVAL_SAMPLES)
            ).read(),
        )
    )

    no_samples = get_samples(
        n=10,
        resolution=dag.final_step.resolution_path.name,
        user_name=user.user_name,
        dag=loaded_dag,
    )
    assert no_samples == {}

    # What happens if source cannot be queried using client?
    # Create new DAG with wrong warehouse (in-memory, no tables)
    bad_dag: DAG = (
        DAG(name=str(dag.name)).load_pending().set_client(sqlite_in_memory_warehouse)
    )

    matchbox_api.get("/eval/samples").mock(
        return_value=Response(200, content=table_to_buffer(samples_no_baz).read())
    )

    with pytest.raises(MatchboxSourceTableError, match="Could not query source"):
        get_samples(
            n=10,
            resolution=dag.final_step.resolution_path.name,
            user_name=user.user_name,
            dag=bad_dag,
        )



"""Composed PostgreSQL adapter for Matchbox server."""

from pydantic import BaseModel
from sqlalchemy import and_, func, or_

from matchbox.common.dtos import ResolutionType
from matchbox.server.base import MatchboxDBAdapter
from matchbox.server.postgresql.adapter.admin import MatchboxPostgresAdminMixin
from matchbox.server.postgresql.adapter.collections import (
    MatchboxPostgresCollectionsMixin,
)
from matchbox.server.postgresql.adapter.eval import MatchboxPostgresEvaluationMixin
from matchbox.server.postgresql.adapter.query import MatchboxPostgresQueryMixin
from matchbox.server.postgresql.db import MBDB, MatchboxPostgresSettings
from matchbox.server.postgresql.orm import (
    Clusters,
    ClusterSourceKey,
    Contains,
    EvalJudgements,
    Probabilities,
    Resolutions,
    SourceConfigs,
)


class FilteredClusters(BaseModel):
    """Wrapper class for filtered cluster queries."""

    has_source: bool | None = None

    def count(self) -> int:
        """Counts the number of clusters in the database."""
        with MBDB.get_session() as session:
            query = session.query(
                func.count(func.distinct(Clusters.cluster_id))
            ).select_from(Clusters)

            if self.has_source is not None:
                if self.has_source:
                    query = query.join(
                        ClusterSourceKey,
                        ClusterSourceKey.cluster_id == Clusters.cluster_id,
                    )
                else:
                    query = (
                        query.join(
                            Probabilities,
                            Probabilities.cluster_id == Clusters.cluster_id,
                            isouter=True,
                        )
                        .join(
                            EvalJudgements,
                            EvalJudgements.endorsed_cluster_id == Clusters.cluster_id,
                            isouter=True,
                        )
                        .filter(
                            or_(
                                EvalJudgements.endorsed_cluster_id.is_not(None),
                                Probabilities.cluster_id.is_not(None),
                            )
                        )
                    )

            return query.scalar()


class FilteredProbabilities(BaseModel):
    """Wrapper class for filtered probability queries."""

    over_truth: bool = False

    def count(self) -> int:
        """Counts the number of probabilities in the database."""
        with MBDB.get_session() as session:
            query = session.query(func.count()).select_from(Probabilities)

            if self.over_truth:
                query = query.join(
                    Resolutions,
                    Probabilities.resolution_id == Resolutions.resolution_id,
                ).filter(
                    and_(
                        Resolutions.truth.isnot(None),
                        Probabilities.probability >= Resolutions.truth,
                    )
                )
            return query.scalar()


class FilteredResolutions(BaseModel):
    """Wrapper class for filtered resolution queries."""

    sources: bool = False
    models: bool = False

    def count(self) -> int:
        """Counts the number of resolutions in the database."""
        with MBDB.get_session() as session:
            query = session.query(func.count()).select_from(Resolutions)

            filter_list = []
            if self.sources:
                filter_list.append(Resolutions.type == ResolutionType.SOURCE)
            if self.models:
                filter_list.append(Resolutions.type == ResolutionType.MODEL)

            if filter_list:
                query = query.filter(or_(*filter_list))

            return query.scalar()


class MatchboxPostgres(
    MatchboxPostgresQueryMixin,
    MatchboxPostgresEvaluationMixin,
    MatchboxPostgresCollectionsMixin,
    MatchboxPostgresAdminMixin,
    MatchboxDBAdapter,
):
    """A PostgreSQL adapter for Matchbox."""

    def __init__(self, settings: MatchboxPostgresSettings) -> None:
        """Initialise the PostgreSQL adapter."""
        self.settings = settings
        MBDB.settings = settings
        MBDB.run_migrations()

        self.sources = SourceConfigs
        self.models = FilteredResolutions(sources=False, models=True)
        self.source_clusters = FilteredClusters(has_source=True)
        self.model_clusters = FilteredClusters(has_source=False)
        self.all_clusters = FilteredClusters()
        self.creates = FilteredProbabilities(over_truth=True)
        self.merges = Contains
        self.proposes = FilteredProbabilities()
        self.source_resolutions = FilteredResolutions(sources=True, models=False)



"""Main CLI entry point for Matchbox."""

import typer
from rich import print

from matchbox import __version__
from matchbox.client.cli import auth, server
from matchbox.client.cli.eval.run import evaluate

app = typer.Typer(
    name="matchbox", help="Matchbox: Entity resolution and data linking framework"
)


@app.command()
def version() -> None:
    """Shows the local Matchbox version."""
    print(f"Matchbox version: {__version__}")


# Add subcommands
app.add_typer(server.app, name="server")
app.add_typer(auth.app, name="auth")
app.command(name="eval")(evaluate)

if __name__ == "__main__":
    app()



"""Remove human resolution type.

Revision ID: 8c7f757b1046
Revises: f500f7d832fe
Create Date: 2025-09-29 15:57:07.548653

"""

from collections.abc import Sequence

from alembic import op

# revision identifiers, used by Alembic.
revision: str = "8c7f757b1046"
down_revision: str | None = "f500f7d832fe"
branch_labels: str | Sequence[str] | None = None
depends_on: str | Sequence[str] | None = None


def upgrade() -> None:
    """Upgrade schema."""
    # Drop the old constraint
    op.drop_constraint(
        "resolution_type_constraints", "resolutions", type_="check", schema="mb"
    )

    # Update existing rows from 'dataset' to 'source'
    op.execute(
        """
        UPDATE mb.resolutions
        SET type = 'source'
        WHERE type = 'dataset'
        """
    )

    # Create the new constraint with only 'model' and 'source'
    op.create_check_constraint(
        "resolution_type_constraints",
        "resolutions",
        "type IN ('model', 'source')",
        schema="mb",
    )


def downgrade() -> None:
    """Downgrade schema."""
    # Drop the new constraint
    op.drop_constraint(
        "resolution_type_constraints", "resolutions", type_="check", schema="mb"
    )

    # Revert 'source' back to 'dataset'
    op.execute(
        """
        UPDATE mb.resolutions
        SET type = 'dataset'
        WHERE type = 'source'
        """
    )
    # Recreate the old constraint with 'model', 'dataset', and 'human'
    op.create_check_constraint(
        "resolution_type_constraints",
        "resolutions",
        "type IN ('model', 'dataset', 'human')",
        schema="mb",
    )



"""Collection and resolution API routes for the Matchbox server."""

import uuid
from typing import Annotated

from fastapi import (
    APIRouter,
    BackgroundTasks,
    Body,
    Depends,
    Query,
    UploadFile,
    status,
)
from pyarrow import ArrowInvalid
from pyarrow import parquet as pq

from matchbox.common.arrow import SCHEMA_INDEX, SCHEMA_RESULTS, table_to_buffer
from matchbox.common.dtos import (
    Collection,
    CollectionName,
    CRUDOperation,
    ErrorResponse,
    ModelResolutionName,
    Resolution,
    ResolutionName,
    ResolutionPath,
    ResolutionType,
    ResourceOperationStatus,
    Run,
    RunID,
    UploadInfo,
)
from matchbox.common.exceptions import (
    MatchboxServerFileError,
)
from matchbox.server.api.dependencies import (
    BackendDependency,
    ParquetResponse,
    SettingsDependency,
    UploadTrackerDependency,
    authorisation_dependencies,
)
from matchbox.server.uploads import file_to_s3, process_upload, process_upload_celery

router = APIRouter(prefix="/collections", tags=["collection"])


# Collection management endpoints


@router.get(
    "",
    summary="List all collections",
    description="Retrieve a list of all collection names in the system.",
)
def list_collections(backend: BackendDependency) -> list[CollectionName]:
    """List all collections."""
    return backend.list_collections()


@router.get(
    "/{collection}",
    responses={404: {"model": ErrorResponse}},
    summary="Get collection details",
    description=(
        "Retrieve details for a specific collection, including all its versions "
        "and resolutions."
    ),
)
def get_collection(
    backend: BackendDependency,
    collection: CollectionName,
) -> Collection:
    """Get collection details with all versions and resolutions."""
    return backend.get_collection(name=collection)


@router.post(
    "/{collection}",
    responses={409: {"model": ErrorResponse}},
    status_code=status.HTTP_201_CREATED,
    dependencies=[Depends(authorisation_dependencies)],
    summary="Create a new collection",
    description="Create a new collection with the specified name.",
)
def create_collection(
    backend: BackendDependency,
    collection: CollectionName,
) -> ResourceOperationStatus:
    """Create a new collection."""
    backend.create_collection(name=collection)
    return ResourceOperationStatus(
        success=True,
        target=f"Collection {collection}",
        operation=CRUDOperation.CREATE,
    )


@router.delete(
    "/{collection}",
    responses={
        404: {"model": ErrorResponse},
        409: {"model": ErrorResponse},
    },
    dependencies=[Depends(authorisation_dependencies)],
    summary="Delete a collection",
    description="Delete a collection and all its versions. Requires confirmation.",
)
def delete_collection(
    backend: BackendDependency,
    collection: CollectionName,
    certain: Annotated[
        bool, Query(description="Confirm deletion of the collection")
    ] = False,
) -> ResourceOperationStatus:
    """Delete a collection."""
    backend.delete_collection(name=collection, certain=certain)
    return ResourceOperationStatus(
        success=True,
        target=f"Collection {collection}",
        operation=CRUDOperation.DELETE,
    )


# Run management endpoints


@router.get(
    "/{collection}/runs/{run_id}",
    responses={404: {"model": ErrorResponse}},
    summary="Get specific run",
    description="Retrieve details for a specific run within a collection.",
)
def get_run(
    backend: BackendDependency,
    collection: CollectionName,
    run_id: RunID,
) -> Run:
    """Get a specific run."""
    return backend.get_run(collection=collection, run_id=run_id)


@router.post(
    "/{collection}/runs",
    responses={404: {"model": ErrorResponse}},
    status_code=status.HTTP_201_CREATED,
    dependencies=[Depends(authorisation_dependencies)],
    summary="Create a new run",
    description="Create a new run within the specified collection.",
)
def create_run(
    backend: BackendDependency,
    collection: CollectionName,
) -> Run:
    """Create a new run in a collection."""
    return backend.create_run(collection=collection)


@router.patch(
    "/{collection}/runs/{run_id}/mutable",
    responses={404: {"model": ErrorResponse}},
    dependencies=[Depends(authorisation_dependencies)],
    summary="Change run mutability",
    description="Set whether a run can be modified.",
)
def set_run_mutable(
    backend: BackendDependency,
    collection: CollectionName,
    run_id: RunID,
    mutable: Annotated[bool, Body(description="Mutability setting")],
) -> ResourceOperationStatus:
    """Set run mutability."""
    backend.set_run_mutable(collection=collection, run_id=run_id, mutable=mutable)
    return ResourceOperationStatus(
        success=True,
        target=f"Run {run_id}",
        operation=CRUDOperation.UPDATE,
    )


@router.patch(
    "/{collection}/runs/{run_id}/default",
    responses={404: {"model": ErrorResponse}},
    dependencies=[Depends(authorisation_dependencies)],
    summary="Change default run",
    description="Set whether a run is the default for its collection.",
)
def set_run_default(
    backend: BackendDependency,
    collection: CollectionName,
    run_id: RunID,
    default: Annotated[bool, Body(description="Default setting")],
) -> ResourceOperationStatus:
    """Set run as default."""
    backend.set_run_default(collection=collection, run_id=run_id, default=default)
    return ResourceOperationStatus(
        success=True,
        target=f"Run {run_id}",
        operation=CRUDOperation.UPDATE,
    )


@router.delete(
    "/{collection}/runs/{run_id}",
    responses={
        404: {"model": ErrorResponse},
        409: {"model": ErrorResponse},
    },
    dependencies=[Depends(authorisation_dependencies)],
    summary="Delete a run",
    description="Delete a run and all its resolutions. Requires confirmation.",
)
def delete_run(
    backend: BackendDependency,
    collection: CollectionName,
    run_id: RunID,
    certain: Annotated[bool, Query(description="Confirm deletion of the run")] = False,
) -> ResourceOperationStatus:
    """Delete a run."""
    backend.delete_run(collection=collection, run_id=run_id, certain=certain)
    return ResourceOperationStatus(
        success=True, target=f"Run {run_id}", operation=CRUDOperation.DELETE
    )


# Resolution management


@router.post(
    "/{collection}/runs/{run_id}/resolutions/{resolution_name}",
    responses={
        404: {"model": ErrorResponse},
        409: {"model": ErrorResponse},
    },
    status_code=status.HTTP_201_CREATED,
    dependencies=[Depends(authorisation_dependencies)],
    summary="Create a resolution",
    description="Create a new resolution (model or source) in the specified run.",
)
def create_resolution(
    backend: BackendDependency,
    collection: CollectionName,
    run_id: RunID,
    resolution_name: ResolutionName,
    resolution: Resolution,
) -> ResourceOperationStatus:
    """Create a resolution (model or source)."""
    resolution_path = ResolutionPath(
        name=resolution_name, collection=collection, run=run_id
    )
    backend.create_resolution(
        resolution=resolution,
        path=resolution_path,
    )
    return ResourceOperationStatus(
        success=True,
        target=f"Resolution {resolution_path}",
        operation=CRUDOperation.CREATE,
    )


@router.get(
    "/{collection}/runs/{run_id}/resolutions/{resolution}",
    responses={404: {"model": ErrorResponse}},
    summary="Get a resolution",
    description="Retrieve a specific resolution (model or source) from the backend.",
)
def get_resolution(
    backend: BackendDependency,
    collection: CollectionName,
    run_id: RunID,
    resolution: ResolutionName,
) -> Resolution:
    """Get a resolution (model or source) from the backend."""
    return backend.get_resolution(
        path=ResolutionPath(collection=collection, run=run_id, name=resolution)
    )


@router.put(
    "/{collection}/runs/{run_id}/resolutions/{resolution_name}",
    responses={404: {"model": ErrorResponse}},
    status_code=status.HTTP_200_OK,
    dependencies=[Depends(authorisation_dependencies)],
    summary="Update a resolution",
    description="Update an existing resolution (model or source) in the specified run.",
)
def update_resolution(
    backend: BackendDependency,
    collection: CollectionName,
    run_id: RunID,
    resolution_name: ResolutionName,
    resolution: Resolution,
) -> ResourceOperationStatus:
    """Update an existing resolution (model or source)."""
    resolution_path = ResolutionPath(
        name=resolution_name, collection=collection, run=run_id
    )
    backend.update_resolution(
        resolution=resolution,
        path=resolution_path,
    )
    return ResourceOperationStatus(
        success=True,
        target=f"Resolution {resolution_path}",
        operation=CRUDOperation.UPDATE,
    )


@router.delete(
    "/{collection}/runs/{run_id}/resolutions/{resolution}",
    responses={
        404: {"model": ErrorResponse},
        409: {"model": ErrorResponse},
    },
    dependencies=[Depends(authorisation_dependencies)],
    summary="Delete a resolution",
    description="Delete a resolution from the backend. Requires confirmation.",
)
def delete_resolution(
    backend: BackendDependency,
    collection: CollectionName,
    run_id: RunID,
    resolution: ResolutionName,
    certain: Annotated[
        bool, Query(description="Confirm deletion of the resolution")
    ] = False,
) -> ResourceOperationStatus:
    """Delete a resolution from the backend."""
    resolution_path = ResolutionPath(collection=collection, run=run_id, name=resolution)
    backend.delete_resolution(path=resolution_path, certain=certain)
    return ResourceOperationStatus(
        success=True,
        target=f"Resolution {resolution_path}",
        operation=CRUDOperation.DELETE,
    )


@router.post(
    "/{collection}/runs/{run_id}/resolutions/{resolution_name}/data",
    responses={
        404: {"model": ErrorResponse},
        400: {"model": ErrorResponse},
        423: {"model": ErrorResponse},
    },
    status_code=status.HTTP_202_ACCEPTED,
    dependencies=[Depends(authorisation_dependencies)],
    summary="Set resolution data",
    description="Create an upload task for source hashes or model results.",
)
def set_data(
    backend: BackendDependency,
    upload_tracker: UploadTrackerDependency,
    settings: SettingsDependency,
    background_tasks: BackgroundTasks,
    collection: CollectionName,
    run_id: RunID,
    resolution_name: ModelResolutionName,
    file: UploadFile,
) -> ResourceOperationStatus:
    """Create an upload task for source hashes or model results."""
    resolution_path = ResolutionPath(
        collection=collection, run=run_id, name=resolution_name
    )

    # Check if data is locked, lock it if not (raises MatchboxLockError -> 423)
    backend.lock_resolution_data(path=resolution_path)

    # Try-except to ensure we release the lock
    try:
        # Validate file extension
        if ".parquet" not in file.filename:
            extension = file.filename.split(".")[-1]
            raise MatchboxServerFileError(
                message=f"Expected .parquet file, got {extension}"
            )

        # pyarrow validates Parquet magic numbers when loading file
        try:
            pq.ParquetFile(file.file)
        except ArrowInvalid as e:
            raise MatchboxServerFileError(
                message=f"Invalid Parquet file: {str(e)}"
            ) from e

        # Get resolution
        resolution = backend.get_resolution(path=resolution_path)

        # Generate unique upload id
        upload_id = str(uuid.uuid4())

        # Upload to S3
        client = backend.settings.datastore.get_client()
        bucket = backend.settings.datastore.cache_bucket_name
        key = f"{upload_id}.parquet"

        if resolution.resolution_type == ResolutionType.SOURCE:
            expected_schema = SCHEMA_INDEX
        else:
            expected_schema = SCHEMA_RESULTS

        table = pq.read_table(file.file)
        if not table.schema.equals(expected_schema):
            raise MatchboxServerFileError(
                message=(
                    "Schema mismatch. "
                    f"Expected:\n{expected_schema}\nGot:\n{table.schema}"
                )
            )

        # Upload to object storage (raises MatchboxServerFileError on failure)
        file_to_s3(client=client, bucket=bucket, key=key, file=file)

        match settings.task_runner:
            case "api":
                background_tasks.add_task(
                    process_upload,
                    backend=backend,
                    tracker=upload_tracker,
                    s3_client=client,
                    resolution_path=resolution_path,
                    upload_id=upload_id,
                    bucket=bucket,
                    filename=key,
                )
            case "celery":
                process_upload_celery.delay(
                    resolution_path_json=resolution_path.model_dump_json(),
                    upload_id=upload_id,
                    bucket=bucket,
                    filename=key,
                )
            case _:
                raise RuntimeError("Unsupported task runner.")

        return ResourceOperationStatus(
            success=True,
            target=f"Resolution data {resolution_path}",
            operation=CRUDOperation.CREATE,
            details=upload_id,
        )
    except:
        backend.unlock_resolution_data(path=resolution_path)
        raise


@router.get(
    "/{collection}/runs/{run_id}/resolutions/{resolution}/data/status",
    responses={404: {"model": ErrorResponse}},
    status_code=status.HTTP_200_OK,
)
def get_upload_status(
    backend: BackendDependency,
    upload_tracker: UploadTrackerDependency,
    collection: CollectionName,
    run_id: RunID,
    resolution: ResolutionName,
    upload_id: Annotated[str | None, Query()] = None,
) -> UploadInfo:
    """Get the status of an upload process.

    Optionally looks for error message if given an upload ID.
    """
    error = None
    if upload_id:
        error = upload_tracker.get(upload_id=upload_id)

    resolution_stage = backend.get_resolution_stage(
        path=ResolutionPath(collection=collection, run=run_id, name=resolution)
    )

    return UploadInfo(stage=resolution_stage, error=error)


@router.get(
    "/{collection}/runs/{run_id}/resolutions/{resolution}/data",
    responses={404: {"model": ErrorResponse}},
    summary="Get resolution results",
    description="Download results for a model as a parquet file.",
)
def get_results(
    backend: BackendDependency,
    collection: CollectionName,
    run_id: RunID,
    resolution: ModelResolutionName,
) -> ParquetResponse:
    """Download results for a model as a parquet file."""
    res = backend.get_model_data(
        path=ResolutionPath(collection=collection, run=run_id, name=resolution)
    )

    buffer = table_to_buffer(res)
    return ParquetResponse(buffer.getvalue())



"""Test the backend adapter's admin functions."""

from functools import partial
from typing import Literal

import pytest
from sqlalchemy import Engine
from test.fixtures.db import BACKENDS

from matchbox.common.dtos import (
    BackendResourceType,
    CollectionName,
    Group,
    GroupName,
    PermissionType,
    User,
)
from matchbox.common.exceptions import (
    MatchboxCollectionNotFoundError,
    MatchboxDataNotFound,
    MatchboxDeletionNotConfirmed,
    MatchboxGroupAlreadyExistsError,
    MatchboxGroupNotFoundError,
    MatchboxSystemGroupError,
    MatchboxUserNotFoundError,
)
from matchbox.common.factories.scenarios import setup_scenario
from matchbox.server.base import MatchboxDBAdapter


@pytest.mark.parametrize("backend", BACKENDS)
@pytest.mark.docker
class TestMatchboxAdminBackend:
    @pytest.fixture(autouse=True)
    def setup(
        self, backend_instance: MatchboxDBAdapter, sqla_sqlite_warehouse: Engine
    ) -> None:
        self.backend: MatchboxDBAdapter = backend_instance
        self.scenario = partial(setup_scenario, warehouse=sqla_sqlite_warehouse)

    # User management

    def test_login_creates_new_user(self) -> None:
        """Login creates a new user if they don't exist."""
        with self.scenario(self.backend, "bare") as _:
            user = User(user_name="alice", email="alice@example.com")
            result = self.backend.login(user)

            assert result.user_name == "alice"
            assert result.email == "alice@example.com"

    def test_login_returns_existing_user(self) -> None:
        """Login returns existing user with same ID."""
        with self.scenario(self.backend, "bare") as _:
            user1 = User(user_name="alice", email="alice@example.com")
            result1 = self.backend.login(user1)

            user2 = User(user_name="alice", email="alice@example.com")
            result2 = self.backend.login(user2)

            assert result1.user_name == result2.user_name

    def test_login_updates_email(self) -> None:
        """Login updates email if it changes."""
        with self.scenario(self.backend, "bare") as _:
            user1 = User(user_name="alice", email="alice@example.com")
            _ = self.backend.login(user1)

            user2 = User(user_name="alice", email="alice@newdomain.com")
            result2 = self.backend.login(user2)

            assert result2.email == "alice@newdomain.com"

    def test_login_different_users(self) -> None:
        """Login creates different IDs for different users."""
        with self.scenario(self.backend, "bare") as _:
            alice = User(user_name="alice")
            bob = User(user_name="bob")

            alice_result = self.backend.login(alice)
            bob_result = self.backend.login(bob)

            assert alice_result.user_name != bob_result.user_name

    # Group management

    def test_create_group(self) -> None:
        """Can create a new group."""
        with self.scenario(self.backend, "bare") as _:
            group = Group(name=GroupName("admins"), description="Administrator group")
            self.backend.create_group(group)

            # Verify it was created
            retrieved = self.backend.get_group(GroupName("admins"))
            assert retrieved.name == "admins"
            assert retrieved.description == "Administrator group"
            assert retrieved.is_system is False
            assert len(retrieved.members) == 0

    def test_create_group_duplicate_fails(self) -> None:
        """Cannot create a group with duplicate name."""
        with self.scenario(self.backend, "bare") as _:
            group1 = Group(name=GroupName("admins"))
            self.backend.create_group(group1)

            group2 = Group(name=GroupName("admins"))
            with pytest.raises(MatchboxGroupAlreadyExistsError):
                self.backend.create_group(group2)

    def test_list_groups_empty(self) -> None:
        """List groups returns empty list when no groups exist."""
        with self.scenario(self.backend, "bare") as _:
            groups = self.backend.list_groups()
            assert groups == []

    def test_list_groups(self) -> None:
        """List groups returns all groups."""
        with self.scenario(self.backend, "bare") as _:
            group1 = Group(name=GroupName("admins"), description="Admins")
            group2 = Group(name=GroupName("users"), description="Users")

            self.backend.create_group(group1)
            self.backend.create_group(group2)

            groups = self.backend.list_groups()
            assert len(groups) == 2
            group_names = {g.name for g in groups}
            assert group_names == {"admins", "users"}

    def test_get_group_not_found(self) -> None:
        """Get group raises error if group doesn't exist."""
        with self.scenario(self.backend, "bare") as _:  # noqa: SIM117
            with pytest.raises(MatchboxGroupNotFoundError):
                self.backend.get_group(GroupName("nonexistent"))

    def test_delete_group(self) -> None:
        """Can delete a group."""
        with self.scenario(self.backend, "bare") as _:
            group = Group(name=GroupName("admins"))
            self.backend.create_group(group)

            # Verify it exists
            retrieved = self.backend.get_group(GroupName("admins"))
            assert retrieved.name == "admins"

            # Delete it
            self.backend.delete_group(GroupName("admins"), certain=True)

            # Verify it's gone
            with pytest.raises(MatchboxGroupNotFoundError):
                self.backend.get_group(GroupName("admins"))

    def test_delete_group_requires_confirmation(self) -> None:
        """Delete group requires certain=True."""
        with self.scenario(self.backend, "bare") as _:
            group = Group(name=GroupName("admins"))
            self.backend.create_group(group)

            with pytest.raises(MatchboxDeletionNotConfirmed):
                self.backend.delete_group(GroupName("admins"), certain=False)

    def test_delete_group_not_found(self) -> None:
        """Delete group raises error if group doesn't exist."""
        with self.scenario(self.backend, "bare") as _:  # noqa: SIM117
            with pytest.raises(MatchboxGroupNotFoundError):
                self.backend.delete_group(GroupName("nonexistent"), certain=True)

    def test_delete_system_group_fails(self) -> None:
        """Cannot delete a system group."""
        with self.scenario(self.backend, "bare") as _:
            system_group = Group(name=GroupName("system_admins"), is_system=True)
            self.backend.create_group(system_group)

            with pytest.raises(MatchboxSystemGroupError):
                self.backend.delete_group(GroupName("system_admins"), certain=True)

    # User-group membership

    def test_add_user_to_group(self) -> None:
        """Can add a user to a group."""
        with self.scenario(self.backend, "bare") as _:
            # Create group
            group = Group(name=GroupName("admins"))
            self.backend.create_group(group)

            # Create user
            user = User(user_name="alice")
            self.backend.login(user)

            # Add user to group
            self.backend.add_user_to_group("alice", GroupName("admins"))

            # Verify membership
            groups = self.backend.get_user_groups("alice")
            assert GroupName("admins") in groups

    def test_add_user_to_nonexistent_user_fails(self) -> None:
        """Cannot add non-existent user to group."""
        with self.scenario(self.backend, "bare") as _:
            # Create group
            group = Group(name=GroupName("admins"))
            self.backend.create_group(group)

            # Try to add non-existent user to group
            with pytest.raises(MatchboxUserNotFoundError):
                self.backend.add_user_to_group("nonexistent", GroupName("admins"))

    def test_add_user_to_group_idempotent(self) -> None:
        """Adding user to group twice doesn't cause error."""
        with self.scenario(self.backend, "bare") as _:
            user = User(user_name="alice", email="alice@example.com")
            self.backend.login(user)

            group = Group(name=GroupName("admins"))
            self.backend.create_group(group)

            self.backend.add_user_to_group("alice", GroupName("admins"))
            self.backend.add_user_to_group("alice", GroupName("admins"))

            groups = self.backend.get_user_groups("alice")
            assert GroupName("admins") in groups

    def test_add_user_to_nonexistent_group_fails(self) -> None:
        """Cannot add user to non-existent group."""
        with self.scenario(self.backend, "bare") as _:
            user = User(user_name="alice", email="alice@example.com")
            self.backend.login(user)

            with pytest.raises(MatchboxGroupNotFoundError):
                self.backend.add_user_to_group("alice", GroupName("nonexistent"))

    def test_remove_user_from_group(self) -> None:
        """Can remove a user from a group."""
        with self.scenario(self.backend, "bare") as _:
            # Setup
            user = User(user_name="alice", email="alice@example.com")
            self.backend.login(user)

            group = Group(name=GroupName("admins"))
            self.backend.create_group(group)
            self.backend.add_user_to_group("alice", GroupName("admins"))

            # Verify user is in group
            groups = self.backend.get_user_groups("alice")
            assert GroupName("admins") in groups

            # Remove user from group
            self.backend.remove_user_from_group("alice", GroupName("admins"))

            # Verify user is no longer in group
            groups = self.backend.get_user_groups("alice")
            assert GroupName("admins") not in groups

    def test_remove_user_from_nonexistent_group_fails(self) -> None:
        """Cannot remove user from non-existent group."""
        with self.scenario(self.backend, "bare") as _:
            user = User(user_name="alice")
            self.backend.login(user)

            with pytest.raises(MatchboxGroupNotFoundError):
                self.backend.remove_user_from_group("alice", GroupName("nonexistent"))

    def test_remove_nonexistent_user_from_group_fails(self) -> None:
        """Cannot remove non-existent user from group."""
        with self.scenario(self.backend, "bare") as _:
            group = Group(name=GroupName("admins"))
            self.backend.create_group(group)

            with pytest.raises(MatchboxUserNotFoundError):
                self.backend.remove_user_from_group("nonexistent", GroupName("admins"))

    def test_get_user_groups_empty(self) -> None:
        """Get user groups returns empty list for user with no groups."""
        with self.scenario(self.backend, "bare") as _:
            user = User(user_name="alice")
            self.backend.login(user)

            groups = self.backend.get_user_groups("alice")
            assert groups == []

    def test_get_user_groups_nonexistent_user_fails(self) -> None:
        """Get user groups raises error for non-existent user."""
        with self.scenario(self.backend, "bare") as _:  # noqa: SIM117
            with pytest.raises(MatchboxUserNotFoundError):
                self.backend.get_user_groups("nonexistent")

    def test_get_user_groups_multiple(self) -> None:
        """Get user groups returns all groups for a user."""
        with self.scenario(self.backend, "bare") as _:
            # Create user
            user = User(user_name="alice", email="alice@example.com")
            self.backend.login(user)

            # Create groups
            group1 = Group(name=GroupName("admins"))
            group2 = Group(name=GroupName("users"))
            self.backend.create_group(group1)
            self.backend.create_group(group2)

            # Add user to both groups
            self.backend.add_user_to_group("alice", GroupName("admins"))
            self.backend.add_user_to_group("alice", GroupName("users"))

            # Verify membership
            groups = self.backend.get_user_groups("alice")
            assert len(groups) == 2
            assert set(groups) == {GroupName("admins"), GroupName("users")}

    def test_get_group_includes_members(self) -> None:
        """Get group returns members list."""
        with self.scenario(self.backend, "bare") as _:
            # Create group and users
            group = Group(name=GroupName("admins"))
            self.backend.create_group(group)

            alice = User(user_name="alice", email="alice@example.com")
            bob = User(user_name="bob", email="bob@example.com")
            self.backend.login(alice)
            self.backend.login(bob)

            # Add users to group
            self.backend.add_user_to_group("alice", GroupName("admins"))
            self.backend.add_user_to_group("bob", GroupName("admins"))

            # Get group and verify members
            retrieved = self.backend.get_group(GroupName("admins"))
            assert len(retrieved.members) == 2
            member_names = {m.user_name for m in retrieved.members}
            assert member_names == {"alice", "bob"}

    # Permissions

    @pytest.mark.parametrize(
        ("resource", "scenario"),
        [
            (BackendResourceType.SYSTEM, "bare"),
            ("collection", "dedupe"),
        ],
        ids=["system", "collection"],
    )
    @pytest.mark.parametrize(
        ("granted_permission", "can_read", "can_write", "can_admin"),
        [
            (PermissionType.READ, True, False, False),
            (PermissionType.WRITE, True, True, False),
            (PermissionType.ADMIN, True, True, True),
        ],
        ids=["read-only", "write-implies-read", "admin-implies-all"],
    )
    def test_permission_hierarchy(
        self,
        resource: Literal[BackendResourceType.SYSTEM] | CollectionName,
        scenario: str,
        granted_permission: PermissionType,
        can_read: bool,
        can_write: bool,
        can_admin: bool,
    ) -> None:
        """Test permission hierarchy: ADMIN > WRITE > READ."""
        with self.scenario(self.backend, scenario) as _:
            group_name = f"{granted_permission}_group"
            user_name = f"user_{granted_permission}"

            # Setup: create group, user, and grant permission
            group = Group(name=GroupName(group_name))
            self.backend.create_group(group)

            user = User(user_name=user_name)
            self.backend.login(user)
            self.backend.add_user_to_group(user_name, GroupName(group_name))

            self.backend.grant_permission(
                GroupName(group_name),
                granted_permission,
                resource,
            )

            # Test: verify permission hierarchy
            assert (
                self.backend.check_permission(user_name, PermissionType.READ, resource)
                == can_read
            )
            assert (
                self.backend.check_permission(user_name, PermissionType.WRITE, resource)
                == can_write
            )
            assert (
                self.backend.check_permission(user_name, PermissionType.ADMIN, resource)
                == can_admin
            )

    @pytest.mark.parametrize(
        ("resource", "scenario"),
        [
            (BackendResourceType.SYSTEM, "bare"),
            ("collection", "dedupe"),
        ],
        ids=["system", "collection"],
    )
    def test_grant_permission(
        self, resource: str | BackendResourceType, scenario: str
    ) -> None:
        """Can grant permission to a group."""
        with self.scenario(self.backend, scenario) as _:
            group = Group(name=GroupName("admins"))
            self.backend.create_group(group)

            self.backend.grant_permission(
                GroupName("admins"),
                PermissionType.ADMIN,
                resource,
            )

            # Verify permission was granted
            permissions = self.backend.get_permissions(resource)
            assert len(permissions) == 1
            assert permissions[0].group_name == "admins"
            assert permissions[0].permission == PermissionType.ADMIN

    @pytest.mark.parametrize(
        ("resource", "scenario"),
        [
            (BackendResourceType.SYSTEM, "bare"),
            ("collection", "dedupe"),
        ],
        ids=["system", "collection"],
    )
    def test_grant_permission_idempotent(
        self, resource: str | BackendResourceType, scenario: str
    ) -> None:
        """Granting same permission twice doesn't cause error."""
        with self.scenario(self.backend, scenario) as _:
            group = Group(name=GroupName("admins"))
            self.backend.create_group(group)

            self.backend.grant_permission(
                GroupName("admins"),
                PermissionType.READ,
                resource,
            )
            self.backend.grant_permission(
                GroupName("admins"),
                PermissionType.READ,
                resource,
            )

            permissions = self.backend.get_permissions(resource)
            assert len(permissions) == 1

    @pytest.mark.parametrize(
        ("resource", "scenario"),
        [
            (BackendResourceType.SYSTEM, "bare"),
            ("collection", "dedupe"),
        ],
        ids=["system", "collection"],
    )
    def test_revoke_permission(
        self, resource: str | BackendResourceType, scenario: str
    ) -> None:
        """Can revoke permission from a group."""
        with self.scenario(self.backend, scenario) as _:
            group = Group(name=GroupName("admins"))
            self.backend.create_group(group)

            # Grant then revoke
            self.backend.grant_permission(
                GroupName("admins"),
                PermissionType.ADMIN,
                resource,
            )
            self.backend.revoke_permission(
                GroupName("admins"),
                PermissionType.ADMIN,
                resource,
            )

            # Verify permission was revoked
            permissions = self.backend.get_permissions(resource)
            assert len(permissions) == 0

    @pytest.mark.parametrize(
        ("resource", "scenario"),
        [
            (BackendResourceType.SYSTEM, "bare"),
            ("collection", "dedupe"),
        ],
        ids=["system", "collection"],
    )
    def test_check_permission_granted(
        self, resource: str | BackendResourceType, scenario: str
    ) -> None:
        """Check permission returns True when user has permission."""
        with self.scenario(self.backend, scenario) as _:
            user = User(user_name="alice", email="alice@example.com")
            self.backend.login(user)

            group = Group(name=GroupName("admins"))
            self.backend.create_group(group)
            self.backend.add_user_to_group("alice", GroupName("admins"))
            self.backend.grant_permission(
                GroupName("admins"),
                PermissionType.READ,
                resource,
            )

            # Check permission
            has_permission = self.backend.check_permission(
                "alice", PermissionType.READ, resource
            )
            assert has_permission is True

    @pytest.mark.parametrize(
        ("resource", "scenario"),
        [
            (BackendResourceType.SYSTEM, "bare"),
            ("collection", "dedupe"),
        ],
        ids=["system", "collection"],
    )
    def test_check_permission_denied(
        self, resource: str | BackendResourceType, scenario: str
    ) -> None:
        """Check permission returns False when user doesn't have permission."""
        with self.scenario(self.backend, scenario) as _:
            user = User(user_name="alice")
            self.backend.login(user)

            # Check permission
            has_permission = self.backend.check_permission(
                "alice", PermissionType.ADMIN, resource
            )
            assert has_permission is False

    def test_check_permission_nonexistent_user(self) -> None:
        """Check permission returns False for non-existent user."""
        with self.scenario(self.backend, "bare") as _:
            has_permission = self.backend.check_permission(
                "nonexistent", PermissionType.READ, BackendResourceType.SYSTEM
            )
            assert has_permission is False

    def test_check_collection_permission_nonexistent_collection(self) -> None:
        """Check permission errors for non-existent collection."""
        with self.scenario(self.backend, "bare") as _:
            user = User(user_name="alice", email="alice@example.com")
            self.backend.login(user)

            group = Group(name=GroupName("readers"))
            self.backend.create_group(group)
            self.backend.add_user_to_group("alice", GroupName("readers"))

            with pytest.raises(MatchboxCollectionNotFoundError):
                _ = self.backend.check_permission(
                    "alice", PermissionType.READ, "nonexistent"
                )

    @pytest.mark.parametrize(
        ("resource", "scenario"),
        [
            (BackendResourceType.SYSTEM, "bare"),
            ("collection", "dedupe"),
        ],
        ids=["system", "collection"],
    )
    def test_get_permissions_empty(
        self, resource: str | BackendResourceType, scenario: str
    ) -> None:
        """Get permissions returns empty list when no permissions granted."""
        with self.scenario(self.backend, scenario) as _:
            permissions = self.backend.get_permissions(resource)
            assert permissions == []

    @pytest.mark.parametrize(
        ("resource", "scenario"),
        [
            (BackendResourceType.SYSTEM, "bare"),
            ("collection", "dedupe"),
        ],
        ids=["system", "collection"],
    )
    def test_get_permissions_multiple_groups(
        self, resource: str | BackendResourceType, scenario: str
    ) -> None:
        """Get permissions returns all permissions from multiple groups."""
        with self.scenario(self.backend, scenario) as _:
            group1 = Group(name=GroupName("readers"))
            group2 = Group(name=GroupName("writers"))
            self.backend.create_group(group1)
            self.backend.create_group(group2)

            self.backend.grant_permission(
                GroupName("readers"),
                PermissionType.READ,
                resource,
            )
            self.backend.grant_permission(
                GroupName("writers"),
                PermissionType.WRITE,
                resource,
            )

            permissions = self.backend.get_permissions(resource)
            assert len(permissions) == 2

            perm_dict = {p.group_name: p.permission for p in permissions}
            assert perm_dict["readers"] == PermissionType.READ
            assert perm_dict["writers"] == PermissionType.WRITE

    def test_permissions_across_multiple_groups(self) -> None:
        """User inherits permissions from all their groups."""
        with self.scenario(self.backend, "dedupe") as _:
            # Create user
            user = User(user_name="alice", email="alice@example.com")
            self.backend.login(user)

            # Create two groups with different permissions
            readers = Group(name=GroupName("readers"))
            writers = Group(name=GroupName("writers"))
            self.backend.create_group(readers)
            self.backend.create_group(writers)

            self.backend.grant_permission(
                GroupName("readers"),
                PermissionType.READ,
                "collection",
            )
            self.backend.grant_permission(
                GroupName("writers"),
                PermissionType.WRITE,
                "collection",
            )

            # Add user to both groups
            self.backend.add_user_to_group("alice", GroupName("readers"))
            self.backend.add_user_to_group("alice", GroupName("writers"))

            # User should have both permissions
            assert self.backend.check_permission(
                "alice", PermissionType.READ, "collection"
            )
            assert self.backend.check_permission(
                "alice", PermissionType.WRITE, "collection"
            )

    # Data management

    def test_validate_ids(self) -> None:
        """Test validating data IDs."""
        with self.scenario(self.backend, "dedupe") as dag_testkit:
            crn_testkit = dag_testkit.sources.get("crn")
            naive_crn_testkit = dag_testkit.models.get("naive_test_crn")

            df_crn = self.backend.query(
                source=crn_testkit.source.resolution_path,
                point_of_truth=naive_crn_testkit.resolution_path,
            )

            ids = df_crn["id"].to_pylist()
            assert len(ids) > 0
            self.backend.validate_ids(ids=ids)

            with pytest.raises(MatchboxDataNotFound):
                self.backend.validate_ids(ids=[-6])

    def test_clear(self) -> None:
        """Test deleting all rows in the database."""
        with self.scenario(self.backend, "dedupe"):
            assert self.backend.sources.count() > 0
            assert self.backend.source_clusters.count() > 0
            assert self.backend.models.count() > 0
            assert self.backend.model_clusters.count() > 0
            assert self.backend.creates.count() > 0
            assert self.backend.merges.count() > 0
            assert self.backend.proposes.count() > 0

            self.backend.clear(certain=True)

            assert self.backend.sources.count() == 0
            assert self.backend.source_clusters.count() == 0
            assert self.backend.models.count() == 0
            assert self.backend.model_clusters.count() == 0
            assert self.backend.creates.count() == 0
            assert self.backend.merges.count() == 0
            assert self.backend.proposes.count() == 0

    def test_clear_and_restore(self) -> None:
        """Test that clearing and restoring the database works."""
        with self.scenario(self.backend, "link") as dag_testkit:
            crn_testkit = dag_testkit.sources.get("crn")
            naive_crn_testkit = dag_testkit.models.get("naive_test_crn")

            count_funcs = [
                self.backend.sources.count,
                self.backend.models.count,
                self.backend.source_clusters.count,
                self.backend.model_clusters.count,
                self.backend.all_clusters.count,
                self.backend.merges.count,
                self.backend.creates.count,
                self.backend.proposes.count,
            ]

            def get_counts() -> list[int]:
                return [f() for f in count_funcs]

            # Verify we have data
            pre_dump_counts = get_counts()
            assert all(count > 0 for count in pre_dump_counts)

            # Get some specific IDs to verify they're restored properly
            df_crn_before = self.backend.query(
                source=crn_testkit.resolution_path,
                point_of_truth=naive_crn_testkit.resolution_path,
            )
            sample_ids_before = df_crn_before["id"].to_pylist()[:5]  # Take first 5 IDs

            # Dump the database
            snapshot = self.backend.dump()

        with self.scenario(self.backend, "bare") as _:
            # Verify counts match pre-dump state
            assert all(c == 0 for c in get_counts())

            # Restore from snapshot
            self.backend.restore(snapshot)

            # Verify counts match pre-dump state
            assert get_counts() == pre_dump_counts

            # Verify specific data was restored correctly
            df_crn_after = self.backend.query(
                source=crn_testkit.resolution_path,
                point_of_truth=naive_crn_testkit.resolution_path,
            )
            sample_ids_after = df_crn_after["id"].to_pylist()[:5]  # Take first 5 IDs

            # The same IDs should be present after restoration
            assert set(sample_ids_before) == set(sample_ids_after)

            # Test that restoring also clears the database
            self.backend.restore(snapshot)

            # Verify counts still match
            assert get_counts() == pre_dump_counts

    def test_delete_orphans(self) -> None:
        """Can delete orphaned clusters."""
        with self.scenario(self.backend, "link") as dag_testkit:
            crn_testkit = dag_testkit.sources.get("crn")
            naive_crn_testkit = dag_testkit.models.get("naive_test_crn")

            # Get number of clusters
            initial_all_clusters = self.backend.all_clusters.count()

            # Delete orphans, none should be deleted yet
            orphans = self.backend.delete_orphans()
            assert orphans == 0
            assert initial_all_clusters == self.backend.all_clusters.count()

            # TODO: insert judgement for cluster, check that it is not deleted when
            # deleting model resolution. Then deleting the judgement should cause
            # exactly 1 orphan.

            model_res = naive_crn_testkit.resolution_path
            self.backend.delete_resolution(model_res, certain=True)

            # Delete orphans, some should be deleted and total clusters should reduce
            orphans = self.backend.delete_orphans()
            assert orphans > 0
            all_clusters_2 = self.backend.all_clusters.count()
            assert initial_all_clusters > all_clusters_2

            # Delete source resolution crn
            source_res = crn_testkit.resolution_path
            self.backend.delete_resolution(source_res, certain=True)

            # Delete orphans again and check number of clusters has reduced
            orphans = self.backend.delete_orphans()
            assert orphans > 0
            all_clusters_3 = self.backend.all_clusters.count()
            assert all_clusters_2 > all_clusters_3



"""Interface to source data."""

from collections.abc import Callable, Generator, Iterable
from functools import wraps
from pathlib import Path
from typing import TYPE_CHECKING, Any, TypeVar, overload

import polars as pl
from pyarrow import Table as ArrowTable
from pyarrow import parquet as pq

from matchbox.client import _handler
from matchbox.client.queries import Query
from matchbox.common.datatypes import DataTypes
from matchbox.common.db import (
    QueryReturnClass,
    QueryReturnType,
)
from matchbox.common.dtos import (
    Resolution,
    ResolutionType,
    SourceConfig,
    SourceField,
    SourceResolutionName,
    SourceResolutionPath,
)
from matchbox.common.exceptions import MatchboxResolutionNotFoundError
from matchbox.common.hash import HashMethod, hash_arrow_table, hash_rows
from matchbox.common.logging import logger, profile_time

if TYPE_CHECKING:
    from matchbox.client.dags import DAG
    from matchbox.client.locations import Location
else:
    DAG = Any
    Location = Any


T = TypeVar("T")


def post_run(method: Callable[..., T]) -> Callable[..., T]:
    """Decorator to ensure that a method is called after source run.

    Raises:
        RuntimeError: If run hasn't happened.
    """

    @wraps(method)
    def wrapper(self: "Source", *args: Any, **kwargs: Any) -> T:
        if self.hashes is None:
            raise RuntimeError(
                "The source must be run before attempting this operation."
            )
        return method(self, *args, **kwargs)

    return wrapper


class Source:
    """Client-side wrapper for source configs."""

    @overload
    def __init__(
        self,
        dag: DAG,
        location: Location,
        name: str,
        extract_transform: str,
        key_field: str,
        index_fields: list[str],
        description: str | None = None,
        infer_types: bool = True,
        validate_etl: bool = True,
    ) -> None: ...

    @overload
    def __init__(
        self,
        dag: DAG,
        location: Location,
        name: str,
        extract_transform: str,
        key_field: SourceField,
        index_fields: list[SourceField],
        description: str | None = None,
        infer_types: bool = False,
        validate_etl: bool = True,
    ) -> None: ...

    def __init__(
        self,
        dag: DAG,
        location: Location,
        name: str,
        extract_transform: str,
        key_field: str | SourceField,
        index_fields: list[str] | list[SourceField],
        description: str | None = None,
        infer_types: bool = False,
        validate_etl: bool = True,
    ):
        """Initialise source.

        Args:
            dag: DAG containing the source.
            location: The location where the source data is stored.
            name: The name of the source.
            description: An optional description of the source.
            extract_transform: The extract/transform logic to apply to the source data.
            key_field: The name of the field to use as the key, or a SourceField
                instance defining the key field. This is the unique identifier we'll
                use to refer to matched data in the source.
            index_fields: The names of the fields to use as index fields, or a list
                of SourceField instances defining the index fields. These are the
                fields you plan to match on.
            infer_types: Whether to infer data types for the fields from the source.
                If False, you must provide SourceField instances for key_field and
                index_fields.
            validate_etl: Whether to skip query validation. If True, it will
                perform query validation. It should be False when loading sources from
                the server. Default True.
        """
        if validate_etl:
            location.validate_extract_transform(extract_transform)

        self.location = location
        self.dag = dag
        self.name = name
        self.description = description
        self.extract_transform = extract_transform
        self.hashes: ArrowTable | None = None

        if infer_types:
            self._validate_fields(key_field, index_fields, str)

            # Assumes client has been set on location
            inferred_types = location.infer_types(extract_transform)
            remote_fields = {
                field_name: SourceField(name=field_name, type=dtype)
                for field_name, dtype in inferred_types.items()
            }
            self.key_field = SourceField(name=key_field, type=DataTypes.STRING)
            self.index_fields = tuple(remote_fields[field] for field in index_fields)
        else:
            self.key_field, self.index_fields = self._validate_fields(
                key_field, index_fields, SourceField
            )

    def _validate_fields(
        self,
        key_field: str | SourceField,
        index_fields: list[str | SourceField],
        type_check: type[str] | type[SourceField],
    ) -> tuple[T, tuple[T, ...]]:
        """Validate that fields match the expected type (str or SourceField)."""
        if not isinstance(key_field, type_check):
            raise ValueError(
                f"Expected {type_check.__name__}, got {type(key_field).__name__}"
            )

        if not all(isinstance(f, type_check) for f in index_fields):
            raise ValueError(
                f"All index_fields must be {type_check.__name__} instances"
            )

        return key_field, tuple(index_fields)

    @property
    def config(self) -> SourceConfig:
        """Generate SourceConfig from Source."""
        return SourceConfig(
            location_config=self.location.config,
            extract_transform=self.extract_transform,
            key_field=self.key_field,
            index_fields=self.index_fields,
        )

    @property
    def sources(self) -> set[SourceResolutionName]:
        """Set of source names upstream of this node."""
        return {self.name}

    @post_run
    def to_resolution(self) -> Resolution:
        """Convert to Resolution for API calls."""
        return Resolution(
            description=self.description,
            truth=None,
            resolution_type=ResolutionType.SOURCE,
            config=self.config,
            fingerprint=hash_arrow_table(self.hashes),
        )

    @classmethod
    def from_resolution(
        cls,
        resolution: Resolution,
        resolution_name: str,
        dag: DAG,
        location: Location,
    ) -> "Source":
        """Reconstruct from Resolution."""
        if resolution.resolution_type != ResolutionType.SOURCE:
            raise ValueError("Resolution must be of type 'source'")

        return cls(
            dag=dag,
            location=location,
            name=SourceResolutionName(resolution_name),
            extract_transform=resolution.config.extract_transform,
            key_field=resolution.config.key_field,
            index_fields=resolution.config.index_fields,
            description=resolution.description,
            infer_types=False,
        )

    def __hash__(self) -> int:
        """Return a hash of the Source based on its config."""
        return hash(self.config)

    def __eq__(self, other: object) -> bool:
        """Check equality of two Source objects based on their config."""
        if not isinstance(other, Source):
            return False
        return self.config == other.config

    @property
    def cache_path(self) -> Path:
        """The path within the DAG cache for storing this source data."""
        return self.dag.cache_path / f"{self.name}.parquet"

    def fetch(
        self,
        qualify_names: bool = False,
        batch_size: int | None = None,
        return_type: QueryReturnType = QueryReturnType.POLARS,
        keys: list[str] | None = None,
    ) -> Generator[QueryReturnClass, None, None]:
        """Applies the extract/transform logic to the source and returns batches lazily.

        Args:
            qualify_names: If True, qualify the names of the columns with the
                source name.
            batch_size: Indicate the size of each batch when fetching data in batches.
            return_type: The type of data to return. Defaults to "polars".
            keys: List of keys to select a subset of all source entries.

        Returns:
            The requested data in the specified format, as an iterator of tables.
        """
        _rename: Callable | None = None
        if qualify_names:

            def _rename(c: str) -> str:
                return self.name + "_" + c

        all_fields = self.config.index_fields + tuple([self.config.key_field])
        schema_overrides = {field.name: field.type.to_dtype() for field in all_fields}

        if keys:
            yield from self.location.execute(
                extract_transform=self.config.extract_transform,
                schema_overrides=schema_overrides,
                rename=_rename,
                batch_size=batch_size,
                return_type=return_type,
                keys=(self.config.key_field.name, keys),
            )
        else:
            yield from self.location.execute(
                extract_transform=self.config.extract_transform,
                schema_overrides=schema_overrides,
                rename=_rename,
                batch_size=batch_size,
                return_type=return_type,
            )

    def sample(
        self, n: int = 100, return_type: QueryReturnType = QueryReturnType.POLARS
    ) -> None:
        """Peek at the top n entries in a source."""
        return next(self.fetch(batch_size=n, return_type=return_type))

    @profile_time(attr="name")
    def run(self, batch_size: int | None = None) -> ArrowTable:
        """Hash a dataset from its warehouse, ready to be inserted, and cache hashes.

        Hashes the index fields defined in the source based on the
        extract/transform logic.

        Does not hash the key field.

        Args:
            batch_size: If set, process data in batches internally. Indicates the
                size of each batch.
        """
        log_prefix = f"Run {self.name}"

        self.cache_path.unlink(missing_ok=True)

        logger.info("Retrieving source data", prefix=log_prefix)
        writer = None
        for batch in self.fetch(
            batch_size=batch_size, return_type=QueryReturnType.ARROW
        ):
            if writer is None:
                # Initialise writer with the first batch's schema
                writer = pq.ParquetWriter(
                    self.cache_path,
                    schema=batch.schema,
                    compression="snappy",
                    use_dictionary=True,
                )

            writer.write_table(batch)  # appends as a new row group

        if writer is not None:
            writer.close()

        key_field: str = self.config.key_field.name
        index_fields: list[str] = [field.name for field in self.config.index_fields]

        all_results: list[pl.DataFrame] = []

        pf = pq.ParquetFile(self.cache_path)
        for rg_idx in range(pf.num_row_groups):
            table = pf.read_row_group(rg_idx)
            batch = pl.from_arrow(table)

            if batch[key_field].is_null().any():
                raise ValueError("keys column contains null values")

            row_hashes: pl.Series = hash_rows(
                df=batch,
                columns=list(sorted(index_fields)),
                method=HashMethod.SHA256,
            )

            result = (
                batch.rename({key_field: "keys"})
                .with_columns(row_hashes.alias("hash"))
                .select(["hash", "keys"])
            )
            all_results.append(result)

        self.hashes = (
            pl.concat(all_results).group_by("hash").agg(pl.col("keys")).to_arrow()
        )

        return self.hashes

    @property
    def resolution_path(self) -> SourceResolutionPath:
        """Returns the source resolution path."""
        return SourceResolutionPath(
            collection=self.dag.name,
            run=self.dag.run,
            name=self.name,
        )

    @property
    def prefix(self) -> str:
        """Get the prefix for the source."""
        return self.config.prefix(self.name)

    @property
    def qualified_key(self) -> str:
        """Get the qualified key for the source."""
        return self.config.qualified_key(self.name)

    @property
    def qualified_index_fields(self) -> list[str]:
        """Get the qualified index fields for the source."""
        return self.config.qualified_index_fields(self.name)

    def qualify_field(self, field: str) -> str:
        """Qualify field names with the source name.

        Args:
            field: The field name to qualify.

        Returns:
            A single qualified field.

        """
        return self.config.qualify_field(self.name, field)

    def f(self, fields: str | Iterable[str]) -> str | list[str]:
        """Qualify one or more field names with the source name.

        Args:
            fields: The field name to qualify, or a list of field names.

        Returns:
            A single qualified field, or a list of qualified field names.

        """
        return self.config.f(self.name, fields)

    @post_run
    @profile_time(attr="name")
    def sync(self) -> None:
        """Send the source config and hashes to the server.

        Not resistant to race conditions: only one client should call sync at a time.
        """
        log_prefix = f"Sync {self.name}"
        resolution = self.to_resolution()
        try:
            existing_resolution = _handler.get_resolution(path=self.resolution_path)
            logger.info("Found existing resolution", prefix=log_prefix)
        except MatchboxResolutionNotFoundError:
            existing_resolution = None

        if existing_resolution:
            if (existing_resolution.fingerprint == resolution.fingerprint) and (
                existing_resolution.config.parents == resolution.config.parents
            ):
                logger.info("Updating existing resolution", prefix=log_prefix)
                # Assumes that resolution hasn't been deleted or made incompatible
                # Else, server will error
                _handler.update_resolution(
                    resolution=resolution, path=self.resolution_path
                )
            else:
                logger.info(
                    "Update not possible. Deleting existing resolution",
                    prefix=log_prefix,
                )
                # Assumes that resolution hasn't been deleted, else server will error
                _handler.delete_resolution(path=self.resolution_path, certain=True)
                existing_resolution = None

        if not existing_resolution:
            logger.info("Creating new resolution", prefix=log_prefix)
            # Assumes that resolution hasn't since been re-created.
            # Else, server will error
            _handler.create_resolution(resolution=resolution, path=self.resolution_path)
            # Assumes resolution has not been deleted or made incompatible
            logger.info("Setting data for new resolution", prefix=log_prefix)
            _handler.set_data(path=self.resolution_path, data=self.hashes)

    def query(self, **kwargs: Any) -> Query:
        """Generate a query for this source."""
        return Query(self, **kwargs, dag=self.dag)

    def clear_data(self) -> None:
        """Deletes data computed for node."""
        self.hashes = None



"""Renamed SourceColumns to SourceFields.

Revision ID: ae63f79f6b39
Revises: e4122bdf9b0d
Create Date: 2025-05-14 12:08:49.377998

"""

from collections.abc import Sequence

from alembic import op

revision: str = "ae63f79f6b39"
down_revision: str | None = "e4122bdf9b0d"
branch_labels: str | Sequence[str] | None = None
depends_on: str | Sequence[str] | None = None


def upgrade() -> None:
    """Rename SourceColumns table and columns to SourceFields."""
    # Drop the old index first
    op.drop_index(
        "ix_source_columns_source_config_id", table_name="source_columns", schema="mb"
    )

    # Drop the foreign key constraint before renaming to control the name
    # PostgreSQL's automatic control was causing issues
    op.drop_constraint(
        "source_columns_source_id_fkey",
        table_name="source_columns",
        type_="foreignkey",
        schema="mb",
    )

    # Rename the table
    op.rename_table(
        old_table_name="source_columns", new_table_name="source_fields", schema="mb"
    )

    # Rename columns to match new naming convention
    op.alter_column(
        table_name="source_fields",
        column_name="column_id",
        new_column_name="field_id",
        schema="mb",
    )

    op.alter_column(
        table_name="source_fields",
        column_name="column_index",
        new_column_name="index",
        schema="mb",
    )

    op.alter_column(
        table_name="source_fields",
        column_name="column_name",
        new_column_name="name",
        schema="mb",
    )

    op.alter_column(
        table_name="source_fields",
        column_name="column_type",
        new_column_name="type",
        schema="mb",
    )

    # Update the unique constraint name
    op.drop_constraint(
        "unique_column_index", table_name="source_fields", type_="unique", schema="mb"
    )

    op.create_unique_constraint(
        "unique_index",
        table_name="source_fields",
        columns=["source_config_id", "index"],
        schema="mb",
    )

    # Recreate the foreign key with the new table/column but keep a name that will
    # revert to the original when we downgrade
    op.create_foreign_key(
        "source_fields_source_config_id_fkey",
        source_table="source_fields",
        referent_table="source_configs",
        local_cols=["source_config_id"],
        remote_cols=["source_config_id"],
        source_schema="mb",
        referent_schema="mb",
        ondelete="CASCADE",
    )

    # Create the index with the same name (it will point to the new table)
    op.create_index(
        "ix_source_columns_source_config_id",
        "source_fields",
        ["source_config_id"],
        unique=False,
        schema="mb",
    )


def downgrade() -> None:
    """Revert SourceFields table and columns back to SourceColumns."""
    # Drop the index
    op.drop_index(
        "ix_source_columns_source_config_id", table_name="source_fields", schema="mb"
    )

    # Drop the foreign key constraint
    op.drop_constraint(
        "source_fields_source_config_id_fkey",
        table_name="source_fields",
        type_="foreignkey",
        schema="mb",
    )

    # Drop unique constraint while we rename columns
    op.drop_constraint(
        "unique_index", table_name="source_fields", type_="unique", schema="mb"
    )

    # Rename columns back to original names
    op.alter_column(
        table_name="source_fields",
        column_name="field_id",
        new_column_name="column_id",
        schema="mb",
    )

    op.alter_column(
        table_name="source_fields",
        column_name="index",
        new_column_name="column_index",
        schema="mb",
    )

    op.alter_column(
        table_name="source_fields",
        column_name="name",
        new_column_name="column_name",
        schema="mb",
    )

    op.alter_column(
        table_name="source_fields",
        column_name="type",
        new_column_name="column_type",
        schema="mb",
    )

    # Rename the table back
    op.rename_table(
        old_table_name="source_fields", new_table_name="source_columns", schema="mb"
    )

    # Create the original index
    op.create_index(
        "ix_source_columns_source_config_id",
        "source_columns",
        ["source_config_id"],
        unique=False,
        schema="mb",
    )

    # Create the unique constraint with the original name
    op.create_unique_constraint(
        "unique_column_index",
        table_name="source_columns",
        columns=["source_config_id", "column_index"],
        schema="mb",
    )

    # Recreate the foreign key constraint
    op.create_foreign_key(
        "source_columns_source_id_fkey",
        source_table="source_columns",
        referent_table="source_configs",
        local_cols=["source_config_id"],
        remote_cols=["source_config_id"],
        source_schema="mb",
        referent_schema="mb",
        ondelete="CASCADE",
    )



"""Interface to locations where source data is stored."""

from __future__ import annotations

from typing import TYPE_CHECKING

from matchbox.common.datatypes import DataTypes

if TYPE_CHECKING:
    from adbc_driver_manager.dbapi import Connection as AdbcConnection

from abc import ABC, abstractmethod
from collections.abc import Callable, Generator, Iterator
from contextlib import contextmanager
from copy import deepcopy
from enum import StrEnum
from functools import wraps
from typing import Any, ParamSpec, Self, TypeVar

import polars as pl
import sqlglot
from sqlalchemy import Connection, Engine
from sqlalchemy.exc import OperationalError
from sqlglot.errors import ParseError

from matchbox.common.db import (
    QueryReturnClass,
    QueryReturnType,
    sql_to_df,
)
from matchbox.common.dtos import LocationConfig, LocationType
from matchbox.common.exceptions import (
    MatchboxSourceClientError,
    MatchboxSourceExtractTransformError,
)
from matchbox.common.logging import logger

T = TypeVar("T")
P = ParamSpec("P")


class ClientType(StrEnum):
    """Enumeration of valid location clients."""

    SQLALCHEMY = "sqlalchemy"
    ADBC = "adbc"


CLIENT_CLASSES: dict[ClientType, type | tuple[type, ...]] = {
    ClientType.SQLALCHEMY: Engine
}

try:
    from adbc_driver_manager.dbapi import Connection as AdbcConnection

    CLIENT_CLASSES[ClientType.ADBC] = AdbcConnection
except ImportError:
    pass


def requires_client(method: Callable[..., T]) -> Callable[..., T]:
    """Decorator that checks if client is set before executing a method.

    A helper method for Location subclasses.

    Raises:
        MatchboxSourceClientError: If the client is not set.
    """

    @wraps(method)
    def wrapper(self: Location, *args: Any, **kwargs: Any) -> T:
        if self.client is None:
            raise MatchboxSourceClientError
        return method(self, *args, **kwargs)

    return wrapper


class Location(ABC):
    """A location for a data source."""

    def __init__(self, name: str) -> None:
        """Initialise location."""
        self.config = LocationConfig(type=self.location_type, name=name)
        self._client = None

    def __deepcopy__(self, memo: dict[int, Any]) -> Self:
        """Create a deep copy of the Location object."""
        obj_copy = type(self)(name=deepcopy(self.config.name, memo))

        # Both objects should share the same client
        if self.client:
            obj_copy.set_client(self.client)

        return obj_copy

    @property
    def client(self) -> Engine | None:
        """Retrieve client."""
        return self._client

    def set_client(self, client: Any) -> Self:  # noqa: ANN401
        """Set client for location and return the location."""
        # Validate against all possible client types
        if not isinstance(client, tuple(CLIENT_CLASSES.values())):
            raise ValueError(
                f"Type {client.__class__} is not valid for {self.__class__}"
            )
        self._client = client
        return self

    @property
    @abstractmethod
    def location_type(self) -> LocationType:
        """Output location type string."""
        ...

    @property
    @abstractmethod
    def client_type(self) -> ClientType:
        """Client type string."""
        ...

    @abstractmethod
    def connect(self) -> bool:
        """Establish connection to the data location.

        Raises:
            AttributeError: If the client is not set.
        """
        ...

    @abstractmethod
    def validate_extract_transform(self, extract_transform: str) -> bool:
        """Validate ET logic against this location's query language.

        Raises:
            MatchboxSourceExtractTransformError: If the ET logic is invalid.
        """
        ...

    @abstractmethod
    def infer_types(self, extract_transform: str) -> dict[str, DataTypes]:
        """Extract all data types from the ET logic."""
        ...

    @abstractmethod
    def execute(
        self,
        extract_transform: str,
        batch_size: int | None = None,
        rename: dict[str, str] | Callable | None = None,
        return_type: QueryReturnType = QueryReturnType.POLARS,
        keys: tuple[str, list[str]] | None = None,
    ) -> Iterator[QueryReturnClass]:
        """Execute ET logic against location and return batches.

        Args:
            extract_transform: The ET logic to execute.
            batch_size: The size used for internal batching. Overrides environment
                variable if set.
            rename: Renaming to apply after the ET logic is executed.

                * If a dictionary is provided, it will be used to rename the columns.
                * If a callable is provided, it will take the old name as input and
                    return the new name.
            return_type: The type of data to return. Defaults to "polars".
            keys: Rule to only retrieve rows by specific keys.
                The key of the dictionary is a field name on which to filter.
                Filters source entries where the key field is in the dict values.

        Raises:
            AttributeError: If the cliet is not set.
        """
        ...

    @classmethod
    def from_config(cls, config: LocationConfig) -> Self:
        """Initialise location from a location config."""
        LocClass = location_type_to_class(config.type)
        return LocClass(name=config.name)


class RelationalDBLocation(Location):
    """A location for a relational database."""

    client: Engine | AdbcConnection | None
    location_type: LocationType = LocationType.RDBMS

    @property
    def client_type(self) -> ClientType | None:
        """Determine client type from the client."""
        if isinstance(self.client, Engine):
            return ClientType.SQLALCHEMY
        elif isinstance(self.client, AdbcConnection):
            return ClientType.ADBC
        return None

    @contextmanager
    def _get_connection(self) -> Generator[Connection | AdbcConnection, None, None]:
        """Context manager for getting database connections with proper cleanup."""
        if self.client_type == ClientType.ADBC:
            # To prevent memory corruption in the ADBC driver when reusing a single
            # connection for multiple queries, we clone the connection
            with self.client.adbc_clone() as connection:
                yield connection
        else:  # SQLAlchemy Engine
            connection = self.client.connect()
            try:
                yield connection
            finally:
                connection.close()

    @requires_client
    def connect(self) -> bool:  # noqa: D102
        try:
            with self._get_connection() as conn:
                _ = sql_to_df(stmt="select 1", connection=conn)
                return True
        except OperationalError:
            return False

    def validate_extract_transform(self, extract_transform: str) -> None:  # noqa: D102
        """Check that the SQL statement only contains a single data-extracting command.

        We are NOT attempting a full sanitisation of the SQL statement
        # Validation is done purely to stop accidental mistakes, not malicious actors
        # Users should only run indexing using SourceConfigs they trust and have read,
        # using least privilege credentials

        Args:
            extract_transform: The SQL statement to validate

        Raises:
            ParseError: If the SQL statement cannot be parsed
            MatchboxSourceExtractTransformError: If validation requirements are not met
        """
        if not extract_transform.strip():
            raise MatchboxSourceExtractTransformError(
                "SQL statement is empty or only contains whitespace."
            )

        # Attempt to make the SQLGlot validation more targeted
        sqlglot_dialect: str | None = None
        client_dialect: str | None = None

        if self.client and self.client_type == ClientType.SQLALCHEMY:
            client_dialect = self.client.dialect.name
        elif self.client and self.client_type == ClientType.ADBC:
            client_dialect = self.client.adbc_get_info().get("vendor_name")
            client_dialect = client_dialect.lower() if client_dialect else None

        match client_dialect:
            case "postgresql":
                sqlglot_dialect = "postgres"
            case "sqlite":
                sqlglot_dialect = "sqlite"
            case _:
                logger.warning("Could not validate specific dialect.")

        try:
            expressions = sqlglot.parse(extract_transform, dialect=sqlglot_dialect)
        except ParseError as e:
            raise MatchboxSourceExtractTransformError(
                "SQL statement could not be parsed."
            ) from e

        if len(expressions) > 1:
            raise MatchboxSourceExtractTransformError(
                "SQL statement contains multiple commands."
            )

        if not expressions:
            raise MatchboxSourceExtractTransformError(
                "SQL statement does not contain any valid expressions."
            )

        expr = expressions[0]

        if not isinstance(expr, sqlglot.expressions.Select | sqlglot.expressions.Union):
            raise MatchboxSourceExtractTransformError(
                "SQL statement must start with a SELECT or WITH command."
            )

        forbidden = (
            sqlglot.expressions.DDL,
            sqlglot.expressions.DML,
            sqlglot.expressions.Into,
        )

        if len(list(expr.find_all(forbidden))) > 0:
            raise MatchboxSourceExtractTransformError(
                "SQL statement must not contain DDL or DML commands."
            )

    @requires_client
    def infer_types(self, extract_transform: str) -> dict[str, DataTypes]:  # noqa: D102
        extract_transform = extract_transform.rstrip(" \t\n;")
        one_row_query = f"select * from ({extract_transform}) as sub limit 1;"
        one_row: pl.DataFrame = next(self.execute(one_row_query))
        column_names = one_row.columns

        inferred_types = {}
        for col in column_names:
            # This expression uses cross-dialect SQL standards;
            # though this is hard to prove as the standard is behind a paywall
            sample_query = (
                f"select {col} from ({extract_transform}) as sub "
                f"where {col} is not null limit 1;"
            )

            sample_row: pl.DataFrame = list(self.execute(sample_query))[0]

            if len(sample_row):
                sample_dtype = sample_row[col].dtype
                inferred_types[col] = DataTypes.from_dtype(sample_dtype)
            else:
                inferred_types[col] = DataTypes.NULL

        return inferred_types

    @requires_client
    def execute(  # noqa: D102
        self,
        extract_transform: str,
        batch_size: int | None = None,
        rename: dict[str, str] | Callable | None = None,
        return_type: QueryReturnType = QueryReturnType.POLARS,
        keys: tuple[str, list[str]] | None = None,
        schema_overrides: dict[str, pl.DataType] | None = None,
    ) -> Generator[QueryReturnClass, None, None]:
        # Strip semicolon, as it can block extended query protocol
        # and slow some engines down
        extract_transform = extract_transform.replace(";", "")

        # We always work in batches to simplify higher level logic
        batch_size: int = batch_size or 10_000

        with self._get_connection() as conn:
            if keys:
                key_field, filter_values = keys
                quoted_key_values = [f"'{key_val}'" for key_val in filter_values]
                # Only filter original SQL if keys are provided
                if quoted_key_values:
                    comma_separated_values = ", ".join(quoted_key_values)
                    # This "IN" expression is a SQL standard;
                    # though this is hard to prove as the standard is behind a paywall
                    extract_transform = (
                        f"select * from ({extract_transform}) as sub "
                        f"where {key_field} in ({comma_separated_values})"
                    )
            yield from sql_to_df(
                stmt=extract_transform,
                schema_overrides=schema_overrides,
                connection=conn,
                rename=rename,
                batch_size=batch_size,
                return_batches=True,
                return_type=return_type,
            )


def location_type_to_class(location_type: LocationType) -> type[Location]:
    """Map location type string to the corresponding class."""
    match location_type:
        case LocationType.RDBMS:
            return RelationalDBLocation
        case _:
            raise ValueError("Location type not recognised.")



from unittest.mock import Mock, patch

import adbc_driver_postgresql.dbapi as adbc_postgres
import polars as pl
import pyarrow as pa
import pytest
from httpx import Response
from polars.testing import assert_frame_equal
from respx import MockRouter
from sqlalchemy import Engine

from matchbox.client.dags import DAG
from matchbox.client.locations import RelationalDBLocation
from matchbox.client.sources import (
    Source,
)
from matchbox.common.datatypes import DataTypes
from matchbox.common.dtos import (
    CRUDOperation,
    ErrorResponse,
    Resolution,
    ResourceOperationStatus,
    SourceField,
    UploadInfo,
    UploadStage,
)
from matchbox.common.exceptions import MatchboxServerFileError
from matchbox.common.factories.sources import (
    source_factory,
    source_from_tuple,
)
from test.fixtures.db import WarehouseConnectionType


def test_source_infers_type(sqla_sqlite_warehouse: Engine) -> None:
    """Creating a source with type inference works."""
    # Create test data
    source_testkit = source_factory(
        n_true_entities=5,
        features=[
            {"name": "name", "base_generator": "word", "datatype": DataTypes.STRING},
        ],
        engine=sqla_sqlite_warehouse,
    ).write_to_location()

    location = RelationalDBLocation(name="dbname").set_client(sqla_sqlite_warehouse)
    source = Source(
        dag=source_testkit.source.dag,
        location=location,
        name="test_source",
        extract_transform=source_testkit.source_config.extract_transform,
        infer_types=True,
        key_field="key",
        index_fields=["name"],
    )

    assert source.config.key_field == SourceField(name="key", type=DataTypes.STRING)
    assert source.config.index_fields == tuple(
        [SourceField(name="name", type=DataTypes.STRING)]
    )


def test_source_sampling_preserves_original_sql(sqla_sqlite_warehouse: Engine) -> None:
    """SQL on RelationalDBLocation is preserved.

    SQLGlot transpiles INSTR() to STR_POSITION() in its default dialect.
    """
    # Create test data
    source_testkit = source_factory(
        n_true_entities=3,
        features=[
            {
                "name": "text_col",
                "base_generator": "word",
                "datatype": DataTypes.STRING,
            },
        ],
        engine=sqla_sqlite_warehouse,
    ).write_to_location()

    location = RelationalDBLocation(name="dbname").set_client(sqla_sqlite_warehouse)

    # Use SQLite's INSTR function (returns position of substring)
    # Other databases use CHARINDEX, POSITION, etc.
    extract_transform = f"""
        SELECT
            key,
            text_col,
            INSTR(text_col, 'a') as position_of_a
        FROM
            "{source_testkit.name}"
    """

    # This should work since INSTR is valid SQLite
    # Would fail if validation transpiles INSTR to POSITION() or similar
    source = Source(
        dag=source_testkit.source.dag,
        location=location,
        name="test_source",
        extract_transform=extract_transform,
        infer_types=True,
        key_field="key",
        index_fields=["text_col", "position_of_a"],
    )

    assert source.config.key_field == SourceField(name="key", type=DataTypes.STRING)
    assert len(source.config.index_fields) == 2

    # This should work if the SQL is preserved exactly
    df = next(source.fetch())
    assert isinstance(df, pl.DataFrame)
    assert len(df) == 3


def test_source_fetch_and_sample(sqla_sqlite_warehouse: Engine) -> None:
    """Test reading source with default parameters."""
    # Create test data
    source_testkit = source_factory(
        n_true_entities=5,
        features=[
            {"name": "name", "base_generator": "word", "datatype": DataTypes.STRING},
        ],
        engine=sqla_sqlite_warehouse,
    ).write_to_location()

    # Create location and source
    location = RelationalDBLocation(name="dbname").set_client(sqla_sqlite_warehouse)
    source = Source(
        dag=source_testkit.source.dag,
        location=location,
        name="test_source",
        extract_transform=source_testkit.source_config.extract_transform,
        infer_types=True,
        key_field="key",
        index_fields=["name"],
    )

    # Execute query
    result = next(source.fetch(batch_size=10))
    assert_frame_equal(result, source.sample(n=10), check_row_order=False)

    # Verify result
    assert isinstance(result, pl.DataFrame)
    assert len(result) == 5
    assert "key" in result.columns
    assert "name" in result.columns

    # Try applying key filter
    key_subset = result[source.config.key_field.name][:2].to_list()
    result = next(source.fetch(keys=key_subset))
    assert len(result) == 2

    # Key filter ineffective with empty list
    result = next(source.fetch(keys=[]))
    assert len(result) == 5


@pytest.mark.parametrize(
    "qualify_names",
    [
        pytest.param(False, id="no_name_qualification"),
        pytest.param(True, id="with_name_qualification"),
    ],
)
@patch("matchbox.client.locations.RelationalDBLocation.execute")
def test_source_fetch_name_qualification(
    mock_execute: Mock, qualify_names: bool, sqlite_in_memory_warehouse: Engine
) -> None:
    """Test that column names are qualified when requested."""
    # Mock the location execute method to verify parameters
    mock_execute.return_value = (x for x in [None])  # execute needs to be a generator
    location = RelationalDBLocation(name="sqlite").set_client(
        sqlite_in_memory_warehouse
    )

    # Create source
    source = Source(
        dag=DAG("collection"),
        location=location,
        name="test_source",
        extract_transform="SELECT key, name FROM users",
        key_field=SourceField(name="key", type=DataTypes.STRING),
        index_fields=[SourceField(name="name", type=DataTypes.STRING)],
    )

    # Call query with qualification parameter
    next(source.fetch(qualify_names=qualify_names))

    # Verify the rename parameter passed to execute
    _, kwargs = mock_execute.call_args
    rename_param = kwargs.get("rename")

    if qualify_names:
        assert rename_param is not None
        assert callable(rename_param)
        # Test the rename function
        sample_col = "test_col"
        assert "test_source_" in source.name + "_" + sample_col
    else:
        assert rename_param is None


@pytest.mark.parametrize(
    ("batch_size", "expected_call_kwargs"),
    [
        pytest.param(
            None,
            {"batch_size": None},
            id="single_return",
        ),
        pytest.param(3, {"batch_size": 3}, id="multiple_batches"),
    ],
)
@patch("matchbox.client.locations.RelationalDBLocation.execute")
def test_source_fetch_batching(
    mock_execute: Mock,
    batch_size: int,
    expected_call_kwargs: dict,
    sqlite_in_memory_warehouse: Engine,
) -> None:
    """Test query with batching options."""
    # Mock the location execute method to verify parameters
    mock_execute.return_value = (x for x in [None])  # execute needs to be a generator
    location = RelationalDBLocation(name="sqlite").set_client(
        sqlite_in_memory_warehouse
    )

    # Create source
    source = Source(
        dag=DAG("collection"),
        location=location,
        name="test_source",
        extract_transform="SELECT key, name FROM users",
        key_field=SourceField(name="key", type=DataTypes.STRING),
        index_fields=[SourceField(name="name", type=DataTypes.STRING)],
    )

    # Call query with batching parameters
    next(source.fetch(batch_size=batch_size))

    # Verify parameters passed to execute
    _, kwargs = mock_execute.call_args
    for key, value in expected_call_kwargs.items():
        assert kwargs.get(key) == value


@pytest.mark.parametrize(
    "batch_size",
    [
        pytest.param(None, id="no_batching"),
        pytest.param(2, id="with_batching"),
    ],
)
def test_source_run(sqla_sqlite_warehouse: Engine, batch_size: int) -> None:
    """Test the run method produces expected hash format."""
    # Create test data with unique values
    n_true_entities = 3
    source_testkit = source_factory(
        n_true_entities=n_true_entities,
        features=[
            {"name": "name", "base_generator": "name", "datatype": DataTypes.STRING},
            {
                "name": "age",
                "base_generator": "random_int",
                "datatype": DataTypes.INT64,
            },
        ],
        engine=sqla_sqlite_warehouse,
    ).write_to_location()

    # Create location and source
    location = RelationalDBLocation(name="dbname").set_client(sqla_sqlite_warehouse)
    source = Source(
        dag=source_testkit.source.dag,
        location=location,
        name="test_source",
        extract_transform=source_testkit.source_config.extract_transform,
        infer_types=True,
        key_field="key",
        index_fields=["name", "age"],
    )

    # Execute run with different batching parameters
    result = source.run(batch_size=batch_size) if batch_size else source.run()

    # Verify result
    assert isinstance(result, pa.Table)
    assert "hash" in result.column_names
    assert "keys" in result.column_names
    assert len(result) == n_true_entities

    source.run()


@patch("matchbox.client.sources.Source.fetch")
def test_source_run_null_identifier(
    mock_fetch: Mock, sqlite_in_memory_warehouse: Engine
) -> None:
    """Test hashing data raises an error when source primary keys contain nulls."""
    # Create a source
    location = RelationalDBLocation(name="sqlite").set_client(
        sqlite_in_memory_warehouse
    )
    source = Source(
        dag=DAG("collection"),
        location=location,
        name="test_source",
        extract_transform="SELECT key, name FROM users",
        key_field=SourceField(name="key", type=DataTypes.STRING),
        index_fields=[SourceField(name="name", type=DataTypes.STRING)],
    )

    # Mock query to return data with null keys
    mock_df = pa.Table.from_pydict({"key": ["1", None], "name": ["a", "b"]})
    mock_fetch.return_value = (x for x in [mock_df])

    # Hashing data should raise ValueErrors for null keys
    with pytest.raises(ValueError, match="keys column contains null values"):
        source.run()


def test_source_sync(matchbox_api: MockRouter, sqla_sqlite_warehouse: Engine) -> None:
    """Test source syncing flow through the API."""
    # Mock source
    testkit = source_factory(
        features=[{"name": "company_name", "base_generator": "company"}],
        engine=sqla_sqlite_warehouse,
    ).write_to_location()

    # Mock the routes:
    # Resolution doesn't yet exist
    matchbox_api.get(
        f"/collections/{testkit.source.dag.name}/runs/{testkit.source.dag.run}/resolutions/{testkit.source.name}"
    ).mock(
        return_value=Response(
            404,
            json=ErrorResponse(
                exception_type="MatchboxResolutionNotFoundError",
                message="Source not found",
            ).model_dump(),
        )
    )
    # Resolution can be inserted
    insert_config_route = matchbox_api.post(
        f"/collections/{testkit.source.dag.name}/runs/{testkit.source.dag.run}/resolutions/{testkit.source.name}"
    ).mock(
        return_value=Response(
            201,
            content=ResourceOperationStatus(
                success=True,
                target=f"Resolution {testkit.source.name}",
                operation=CRUDOperation.CREATE,
            ).model_dump_json(),
        )
    )
    # Resolution data can be inserted
    insert_hashes_route = matchbox_api.post(
        f"/collections/{testkit.source.dag.name}/runs/{testkit.source.dag.run}/resolutions/{testkit.source.name}/data"
    ).mock(
        return_value=Response(
            202,
            json=ResourceOperationStatus(
                success=True, target="", operation=CRUDOperation.CREATE
            ).model_dump(),
        )
    )

    # Later, resolution can be updated
    update_route = matchbox_api.put(
        f"/collections/{testkit.source.dag.name}/runs/{testkit.source.dag.run}/resolutions/{testkit.source.name}"
    ).mock(
        return_value=Response(
            200,
            content=ResourceOperationStatus(
                success=True,
                target=f"Resolution {testkit.source.name}",
                operation=CRUDOperation.UPDATE,
            ).model_dump_json(),
        )
    )

    # Later, resolution can be deleted and recreated
    delete_route = matchbox_api.delete(
        f"/collections/{testkit.source.dag.name}/runs/{testkit.source.dag.run}/resolutions/{testkit.source.name}"
    ).mock(
        return_value=Response(
            200,
            content=ResourceOperationStatus(
                success=True,
                target=f"Resolution {testkit.source.name}",
                operation=CRUDOperation.DELETE,
            ).model_dump_json(),
        )
    )

    # -- ERRORS --

    # Can't sync before running
    with pytest.raises(RuntimeError, match="must be run"):
        testkit.source.sync()

    # We now run, but test that upload failure is handled
    testkit.fake_run()
    # If stage is not PROCESSING, job will fail
    matchbox_api.get(
        f"/collections/{testkit.source.dag.name}/runs/{testkit.source.dag.run}/resolutions/{testkit.source.name}/data/status"
    ).mock(
        return_value=Response(
            200, json=UploadInfo(stage=UploadStage.READY).model_dump()
        )
    )
    with pytest.raises(MatchboxServerFileError, match="problem"):
        testkit.source.sync()

    # -- FIRST TIME INSERTION --

    # Before upload, resolution is ready for data, after it is complete
    matchbox_api.get(
        f"/collections/{testkit.source.dag.name}/runs/{testkit.source.dag.run}/resolutions/{testkit.source.name}/data/status"
    ).mock(
        return_value=Response(
            200, json=UploadInfo(stage=UploadStage.COMPLETE).model_dump()
        )
    )

    # Sync the source, successfully
    testkit.source.run()
    testkit.source.sync()
    # Source was created, not updated or deleted
    assert insert_config_route.called
    assert not update_route.called
    assert not delete_route.called
    # Resolution metadata was correct
    resolution_call = Resolution.model_validate_json(
        insert_config_route.calls.last.request.content.decode("utf-8")
    )
    assert resolution_call == testkit.source.to_resolution()
    # Resolution data was correct
    assert (
        b"Content-Disposition: form-data;"
        in insert_hashes_route.calls.last.request.content
    )
    assert b"PAR1" in insert_hashes_route.calls.last.request.content

    # -- SOFT UPDATE --

    insert_hashes_route.reset()
    # Mock endpoint now returns existing resolution
    matchbox_api.get(
        f"/collections/{testkit.source.dag.name}/runs/{testkit.source.dag.run}/resolutions/{testkit.source.name}"
    ).mock(return_value=Response(200, json=testkit.source.to_resolution().model_dump()))

    # Mock endpoint now declares data is present already
    matchbox_api.get(
        f"/collections/{testkit.source.dag.name}/runs/{testkit.source.dag.run}/resolutions/{testkit.source.name}/data/status"
    ).mock(
        return_value=Response(
            200, json=UploadInfo(stage=UploadStage.COMPLETE).model_dump()
        )
    )
    testkit.source.sync()
    # Resolution was compatible: ensure it was updated, not deleted
    assert update_route.called
    assert not delete_route.called
    # The data did not need to be updated
    assert not insert_hashes_route.called

    # -- HARD UPDATE --

    insert_hashes_route.reset()
    # Mock endpoint now returns existing resolution
    matchbox_api.get(
        f"/collections/{testkit.source.dag.name}/runs/{testkit.source.dag.run}/resolutions/{testkit.source.name}"
    ).mock(return_value=Response(200, json=testkit.source.to_resolution().model_dump()))

    # Changing data requires deletion and re-insertion
    testkit.data_hashes = testkit.data_hashes.slice(1, 3)
    testkit.fake_run()

    # Resolution data is first ready to upload, and then uploaded
    matchbox_api.get(
        f"/collections/{testkit.source.dag.name}/runs/{testkit.source.dag.run}/resolutions/{testkit.source.name}/data/status"
    ).mock(return_value=Response(200, json=testkit.source.to_resolution().model_dump()))

    testkit.source.sync()

    assert delete_route.called
    assert insert_hashes_route.called


@pytest.mark.docker
@pytest.mark.parametrize("warehouse", ["adbc_postgres", "sqla_postgres"], indirect=True)
@pytest.mark.parametrize(
    ("mb_type", "pl_type", "test_data", "expected_value"),
    [
        pytest.param(
            DataTypes.LIST(DataTypes.STRING),
            pl.List(pl.String),
            ({"tags": ["a", "b"]}, {"tags": ["c", "d"]}),
            ["a", "b"],
            id="List",
        ),
        pytest.param(
            DataTypes.ARRAY(DataTypes.STRING, shape=2),
            pl.Array(pl.String, shape=2),
            ({"tags": ["a", "b"]}, {"tags": ["c", "d"]}),
            ["a", "b"],
            id="Array",
        ),
        # Structs not supported in PostgreSQL
    ],
)
def test_source_fetch_complex_types_postgres(
    warehouse: WarehouseConnectionType,
    adbc_postgres_warehouse: adbc_postgres.Connection,
    mb_type: DataTypes,
    pl_type: pl.DataType,
    test_data: tuple[dict, dict],
    expected_value: list | dict,
) -> None:
    """Test that Source handles complex types correctly with a real Postgres DB."""
    keys = ("1", "2")

    # Write test data to Postgres
    source_testkit = source_from_tuple(
        data_tuple=test_data,
        data_keys=keys,
        name="test_complex_source_pg",
        engine=adbc_postgres_warehouse,
    ).write_to_location()

    # Configure Source with explicit type
    location = RelationalDBLocation(name="postgres_db").set_client(warehouse)

    source = Source(
        dag=source_testkit.source.dag,
        location=location,
        name=source_testkit.source.name,
        extract_transform=source_testkit.source.config.extract_transform,
        infer_types=False,
        key_field=source_testkit.source.config.key_field,
        index_fields=[SourceField(name="tags", type=mb_type)],
    )

    # Fetch and verify
    results = list(source.fetch())
    df = results[0]

    assert len(df) == 2
    assert df["tags"].dtype == pl_type

    if isinstance(expected_value, list):
        assert df["tags"][0].to_list() == expected_value
    elif isinstance(expected_value, dict):
        assert dict(df["tags"][0]) == expected_value



"""Utilities for hashing data and creating unique identifiers."""

import base64
import hashlib
from enum import StrEnum
from typing import TypeVar
from uuid import UUID

import polars as pl
import polars.expr as plx
import polars_hash as plh
import pyarrow as pa

T = TypeVar("T")
HashableItem = TypeVar("HashableItem", bytes, bool, str, int, float, bytearray)

HASH_FUNC = hashlib.sha256


class HashMethod(StrEnum):
    """Supported hash methods for row hashing."""

    XXH3_128 = "xxh3_128"
    SHA256 = "sha256"


def hash_to_base64(hash: bytes) -> str:
    """Converts a hash to a base64 string."""
    return base64.urlsafe_b64encode(hash).decode("utf-8")


def base64_to_hash(value: str | bytes) -> bytes:
    """Converts a base64 string to a hash, or returns a hash as is."""
    if isinstance(value, bytes):
        return value
    return base64.urlsafe_b64decode(value)


def prep_for_hash(item: HashableItem) -> bytes:
    """Encodes strings so they can be hashed, otherwises, passes through."""
    if isinstance(item, bytes):
        return item
    elif isinstance(item, str):
        return bytes(item.encode())
    elif isinstance(item, UUID):
        return item.bytes
    elif isinstance(item, int):
        # https://stackoverflow.com/a/54141411
        signed = True
        length = ((item + ((item * signed) < 0)).bit_length() + 7 + signed) // 8
        return item.to_bytes(length, byteorder="big", signed=signed)
    else:
        raise ValueError(f"Cannot hash value of type {type(item)}")


def hash_data(data: HashableItem) -> bytes:
    """Hash the given data using the globally defined hash function.

    This function ties into the existing hashing utilities.
    """
    return HASH_FUNC(prep_for_hash(data)).digest()


def hash_values(*values: tuple[T, ...]) -> bytes:
    """Returns a single hash of a tuple of items ordered by its values.

    List must be sorted as the different orders of value must produce the same hash.
    """
    try:
        sorted_vals = sorted(values)
    except TypeError as e:
        raise TypeError("Can only order lists or fields of the same datatype.") from e

    hashed_vals_list = [HASH_FUNC(prep_for_hash(i)) for i in sorted_vals]

    hashed_vals = hashed_vals_list[0]
    for val in hashed_vals_list[1:]:
        hashed_vals.update(val.digest())

    return hashed_vals.digest()


def process_column_for_hashing(column_name: str, schema_type: pl.DataType) -> plx.Expr:
    """Process a column for hashing based on its type.

    Args:
        column_name: The column name
        schema_type: The polars schema type of the column

    Returns:
        A polars expression for processing the column
    """
    if isinstance(schema_type, pl.Binary):
        return (
            pl.col(column_name).fill_null("\x00").bin.encode("hex").alias(column_name)
        )
    elif isinstance(schema_type, pl.Struct):
        return (
            pl.col(column_name)
            .struct.json_encode()
            .fill_null("\x00")
            .alias(column_name)
        )
    elif isinstance(schema_type, pl.List):
        return pl.col(column_name).list.join(",").fill_null("\x00").alias(column_name)
    else:
        return pl.col(column_name).cast(pl.Utf8).fill_null("\x00").alias(column_name)


def hash_rows(
    df: pl.DataFrame, columns: list[str], method: HashMethod = HashMethod.XXH3_128
) -> pl.Series:
    """Hash all rows in a dataframe.

    Args:
        df: The DataFrame to hash rows from
        columns: The column names to include in the hash
        method: The hash method to use

    Returns:
        List of row hashes as bytes
    """
    expr_list = [
        process_column_for_hashing(column, df.schema[column]) for column in columns
    ]
    df_processed = df.with_columns(expr_list)

    record_separator = "␞"
    unit_separator = "␟"

    str_concatenation: list[pl.Expr] = []
    for c in columns:
        str_concatenation.extend(
            [
                pl.lit(c),  # column name
                pl.lit(unit_separator),
                pl.col(c),  # column value
                pl.lit(record_separator),
            ]
        )

    if method == HashMethod.XXH3_128:
        row_hashes = df_processed.select(
            plh.concat_str(str_concatenation).nchash.xxh3_128().alias("row_hash")
        )
        return row_hashes["row_hash"]
    elif method == HashMethod.SHA256:
        row_hashes = df_processed.select(
            plh.concat_str(str_concatenation)
            .chash.sha2_256()
            .str.decode("hex")
            .alias("row_hash")
        )
        return row_hashes["row_hash"]
    else:
        raise ValueError(f"Unsupported hash method: {method}")


def hash_arrow_table(
    table: pa.Table,
    method: HashMethod = HashMethod.XXH3_128,
    as_sorted_list: list[str] | None = None,
) -> bytes:
    """Computes a content hash of an Arrow table invariant to row and field order.

    This is used to content-address an Arrow table for caching.

    Args:
        table: The pyarrow Table to hash
        method: The method to use for hashing rows (XXH3_128 or SHA256)
        as_sorted_list: Optional list of column names to hash as a sorted list.
            For example, ["left_id", "right_id"] will create a "sorted_list"
            column and drop the original columns to ensure (1,2) and (2,1)
            hash to the same value. Works with 2 or more columns.

            Note: if list columns are combined with a column that's nullable,
            list + null value returns null. See Polars' concat_list documentation
            for more details.

    Returns:
        Bytes representing the content hash of the table
    """
    df = pl.from_arrow(table)

    if df.height == 0:
        return b"empty_table_hash"

    # Apply normalisation if specified
    if as_sorted_list:
        if len(as_sorted_list) < 2:
            raise ValueError(
                "Lists passed to as_sorted_list must contain at least 2 column names"
            )

        # Check that all columns exist
        missing_cols = [col for col in as_sorted_list if col not in df.columns]
        if missing_cols:
            raise ValueError(f"Columns not found in dataframe: {missing_cols}")

        # Create normalised group and drop original columns
        df = df.with_columns(
            pl.concat_list(as_sorted_list).list.sort().alias("sorted_list")
        ).drop(as_sorted_list)

    columns: list[str] = sorted(df.columns)
    df = df.select(columns)

    # Explode list fields
    for column in columns:
        if isinstance(df.schema[column], pl.List):
            df = df.explode(column)

    df = df.sort(by=columns)
    row_hashes = hash_rows(df=df, columns=columns, method=method)
    all_hashes: bytes = b"".join(row_hashes.sort().to_list())

    return HASH_FUNC(all_hashes).digest()


def hash_model_results(results: pa.Table) -> bytes:
    """Fingerprint model results."""
    return hash_arrow_table(results, as_sorted_list=["left_id", "right_id"])


class IntMap:
    """A data structure to map integers without collisions within a dedicated space.

    A stand-in for hashing integers within pa.int64.

    Takes unordered sets of integers, and maps them a to an ID that
    1) is a negative integer; 2) does not collide with other IDs generated by other
    instances of this class, as long as they are initialised with a different salt.

    The fact that IDs are always negative means that it's possible to build a hierarchy
    where IDs are themselves parts of other sets, and it's easy to distinguish integers
    mapped to raw data points (which will be non-negative), to integers that are IDs
    (which will be negative). The salt allows to work with a parallel execution
    model, where each worker maintains their separate ID space, as long as each worker
    operates on disjoint subsets of positive integers.

    Args:
        salt (optional): A positive integer to salt the Cantor pairing function
    """

    def __init__(self, salt: int = 42) -> None:
        """Initialise the IntMap."""
        self.mapping: dict[frozenset[int], int] = {}
        if salt < 0:
            raise ValueError("The salt must be a positive integer")
        self.salt: int = salt

    def _salt_id(self, i: int) -> int:
        """Use Cantor pairing function on the salt and a negative int ID.

        It negates the Cantor pairing function to always return a negative integer.
        """
        if i >= 0:
            raise ValueError("ID must be a negative integer")
        return -int(0.5 * (self.salt - i) * (self.salt - i + 1) - i)

    def index(self, *values: int) -> int:
        """Index a set of integers.

        Args:
            values: the integers in the set you want to index

        Returns:
            The old or new ID corresponding to the set
        """
        value_set = frozenset(values)
        if value_set in self.mapping:
            return self.mapping[value_set]

        new_id: int = -len(self.mapping) - 1
        salted_id = self._salt_id(new_id)
        self.mapping[value_set] = salted_id

        return salted_id

    def has_mapping(self, *values: int) -> bool:
        """Check if index for values already exists.

        Args:
            values: the integers in the set you want to index

        Returns:
            Boolean indicating whether index for values already exists
        """
        value_set = frozenset(values)
        return value_set in self.mapping



"""Add tag to judgement.

Revision ID: 58ee4e19fb7d
Revises: 13095b44ff09
Create Date: 2025-12-05 16:59:15.778813

"""

from collections.abc import Sequence

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision: str = "58ee4e19fb7d"
down_revision: str | None = "13095b44ff09"
branch_labels: str | Sequence[str] | None = None
depends_on: str | Sequence[str] | None = None


def upgrade() -> None:
    """Upgrade schema."""
    op.add_column(
        "eval_judgements", sa.Column("tag", sa.TEXT(), nullable=True), schema="mb"
    )


def downgrade() -> None:
    """Downgrade schema."""
    op.drop_column("eval_judgements", "tag", schema="mb")



import logging
from collections.abc import Generator

import pytest
from adbc_driver_manager import AdbcConnection
from httpx import Client
from polars.testing import assert_frame_equal

from matchbox.client.dags import DAG
from matchbox.client.locations import RelationalDBLocation
from matchbox.client.models.dedupers import NaiveDeduper
from matchbox.client.models.linkers import DeterministicLinker
from matchbox.client.sources import Source, SourceField
from matchbox.common.datatypes import DataTypes
from matchbox.common.exceptions import MatchboxResolutionNotFoundError
from matchbox.common.factories.sources import (
    FeatureConfig,
    SourceTestkitParameters,
    SuffixRule,
    linked_sources_factory,
)


@pytest.mark.docker
class TestE2EPipelineBuilder:
    """End to end tests for DAG pipeline functionality."""

    def _clean_company_name(self, column: str) -> str:
        """Generate cleaning SQL for a company name column.

        Removes company suffixes (Ltd, Limited) and normalises whitespace
        from the company_name field.
        """

        return f"""
            trim(
                regexp_replace(
                    regexp_replace(
                        {column},
                        ' (Ltd|Limited)$',
                        '',
                        'g'
                    ),
                    '\\s+',
                    ' ',
                    'g'
                )
            )
        """

    @pytest.fixture(autouse=True, scope="function")
    def setup_environment(
        self,
        matchbox_client: Client,
        adbc_postgres_warehouse: AdbcConnection,
    ) -> Generator[None, None, None]:
        """Set up warehouse and database using fixtures."""
        # Persist shared setup for use in the test body
        n_true_entities = 10  # Keep it small for simplicity
        self.warehouse_engine = adbc_postgres_warehouse

        # Create simple feature configurations - just two sources
        features = {
            "company_name": FeatureConfig(
                name="company_name",
                base_generator="company",
            ),
            "registration_id": FeatureConfig(
                name="registration_id",
                base_generator="bothify",
                parameters=(("text", "REG-###-???"),),
            ),
            "tags": FeatureConfig(
                name="tags",
                base_generator="words",
                parameters=(("nb", 3),),
            ),
        }

        # Create two simple sources that can be linked
        source_parameters = (
            SourceTestkitParameters(
                name="source_a",
                engine=self.warehouse_engine,
                features=(
                    features["company_name"].add_variations(
                        SuffixRule(suffix=" Ltd"),
                        SuffixRule(suffix=" Limited"),
                    ),
                    features["registration_id"],
                ),
                n_true_entities=n_true_entities,
                repetition=0,  # No duplicates
            ),
            SourceTestkitParameters(
                name="source_b",
                engine=self.warehouse_engine,
                features=(
                    features["company_name"],
                    features["registration_id"],
                    features["tags"],
                ),
                n_true_entities=n_true_entities // 2,  # Half overlap
                repetition=1,  # Duplicate all rows for deduplication testing
            ),
        )

        # Create linked sources testkit
        self.linked_testkit = linked_sources_factory(
            source_parameters=source_parameters,
            seed=42,
        )

        # Setup - Create tables in warehouse
        for source_testkit in self.linked_testkit.sources.values():
            source_testkit.write_to_location()

        # Clear matchbox database before test
        response = matchbox_client.delete("/database", params={"certain": "true"})
        assert response.status_code == 200, "Failed to clear matchbox database"

        yield

        # Teardown
        response = matchbox_client.delete("/database", params={"certain": "true"})
        assert response.status_code == 200, "Failed to clear matchbox database"

    def test_dag_pipeline_creation_and_rerun(self) -> None:
        """Test DAG API with simple two-source pipeline

        Rerun to test overwriting.
        """

        # === SETUP PHASE ===
        dw_loc = RelationalDBLocation(name="dbname").set_client(self.warehouse_engine)

        dag = DAG("companies").new_run()

        # Create source configs
        source_a = dag.source(
            location=dw_loc,
            name="source_a",
            extract_transform="""
                select
                    key::text as id,
                    company_name,
                    registration_id
                from
                    source_a;
            """,
            infer_types=True,
            key_field="id",
            index_fields=["company_name", "registration_id"],
        )

        source_b = dag.source(
            location=dw_loc,
            name="source_b",
            extract_transform="""
                select
                    key::text as id,
                    company_name,
                    registration_id,
                    tags::text[] as tags
                from
                    source_b;
            """,
            infer_types=True,
            key_field="id",
            index_fields=["company_name", "registration_id", "tags"],
        )

        # Dedupe steps
        dedupe_a = source_a.query(
            cleaning={
                "company_name": self._clean_company_name(source_a.f("company_name")),
                "registration_id": source_a.f("registration_id"),
            },
        ).deduper(
            name="dedupe_source_a",
            description="Deduplicate source A",
            model_class=NaiveDeduper,
            model_settings={"unique_fields": ["registration_id"]},
        )

        dedupe_b = source_b.query(
            cleaning={
                "company_name": self._clean_company_name(source_b.f("company_name")),
                "registration_id": source_b.f("registration_id"),
            }
        ).deduper(
            name="dedupe_source_b",
            description="Deduplicate source B",
            model_class=NaiveDeduper,
            model_settings={"unique_fields": ["registration_id"]},
        )

        # Link deduplicated sources A and B
        link_a_b = dedupe_a.query(
            source_a,
            cleaning={
                "company_name": self._clean_company_name(source_a.f("company_name")),
                "registration_id": source_a.f("registration_id"),
            },
        ).linker(
            dedupe_b.query(
                source_b,
                cleaning={
                    "company_name": self._clean_company_name(
                        source_b.f("company_name")
                    ),
                    "registration_id": source_b.f("registration_id"),
                },
            ),
            name="final",
            description="Link sources A and B on registration ID",
            model_class=DeterministicLinker,
            model_settings={"comparisons": ["l.registration_id = r.registration_id"]},
        )

        # === FIRST RUN ===
        logging.info("Running DAG for the first time")
        dag.run_and_sync()

        assert DAG.list_all() == [dag.name]

        # Update metadata of one node, will check later
        link_a_b.description = "Updated description"
        link_a_b.sync()

        # Basic verification - we have some linked results and can retrieve them
        final_df = link_a_b.query(source_a, source_b).data()

        # # Should have linked results
        assert len(final_df) > 0, "Expected some results from first run"
        assert final_df["id"].n_unique() == len(self.linked_testkit.true_entities)

        first_run_entities = final_df["id"].n_unique()
        logging.info(f"First run produced {first_run_entities} unique entities")

        # Lookup works too
        test_key = next(
            iter(
                self.linked_testkit.find_entities(
                    min_appearances={source_b.name: 1, source_a.name: 1}
                )[0].keys[source_b.name]
            )
        )

        matches = dag.lookup_key(
            from_source=source_b.name,
            to_sources=[source_a.name],
            key=test_key,
        )
        assert len(matches[source_a.name]) >= 1

        # Can retrieve whole lookup
        dag1_lookup = dag.resolve().as_lookup()

        # Set as new default
        dag.set_default()

        # === SECOND RUN ===

        logging.info("Running DAG again to test downloading and using the default")

        # Load default
        reconstructed_dag = DAG("companies").load_default()
        assert reconstructed_dag.run == dag.run

        # Check complex types serialise and deserialise
        source_b_remote: Source = reconstructed_dag.get_source("source_b")
        tags_field: SourceField = next(
            f for f in source_b_remote.index_fields if f.name == "tags"
        )
        assert tags_field.type == DataTypes.LIST(DataTypes.STRING)

        # Previous update was effective
        assert reconstructed_dag.get_model("final").description == "Updated description"

        # Can directly read data from default
        assert matches == reconstructed_dag.lookup_key(
            from_source=source_b.name,
            to_sources=[source_a.name],
            key=test_key,
        )

        # Start a new run
        rerun_dag = reconstructed_dag.set_client(self.warehouse_engine).new_run()
        assert rerun_dag.run != dag.run
        rerun_dag.run_and_sync()

        # The lookup is identical
        assert_frame_equal(
            rerun_dag.resolve().as_lookup(),
            dag1_lookup,
            check_column_order=False,
            check_row_order=False,
        )

        # Load pending to check we can
        pending_dag: DAG = (
            DAG("companies").load_pending().set_client(self.warehouse_engine)
        )
        assert pending_dag.run == rerun_dag.run

        logging.info("DAG pipeline test completed successfully!")

        # Possible to overwrite node locally
        # (following source has one fewer field)
        source_a = pending_dag.source(
            location=dw_loc,
            name="source_a",
            extract_transform="""
                select
                    key::text as id,
                    company_name
                from
                    source_a;
            """,
            infer_types=True,
            key_field="id",
            index_fields=["company_name"],
        )

        source_a.run()
        # Possible to overwrite one node on server
        source_a.sync()
        # This will cause downstream queries to fail
        with pytest.raises(MatchboxResolutionNotFoundError):
            pending_dag.get_model("final").query(source_a).data()



"""Modal screens for entity resolution evaluation."""

from textual import events, on
from textual.app import ComposeResult
from textual.containers import Container
from textual.screen import ModalScreen
from textual.widgets import Button, Static

HELP_TEXT = """
A model thinks this data is one entity. Is it?

[b]Paint the columns to label this as one or more entities.[/b]

• Each column is data from a single source.
• Each row is a field from that source.

If you think the model got it right, paint all columns the same.

If you think column 1 and 2 belong to one entity, and 3-5 another,
paint them that way.

When you're done, submit, and do another!

• Press letter (a-z): Select that group (26 groups available)
• Press number (1-9, 0): Assign column to selected group
* Space: Submit completed review
• Esc: Clear group selection
• Shift+Right: Skip to next entity
• Left/Right: Switch between tabs
• Up/Down: Page through rows
""".strip()

NO_SAMPLES_TEXT = """
No samples are available for this resolution.

Possible reasons:
• All clusters have been recently judged by you
• The resolution has no probability data
• No clusters exist for this resolution

Press Ctrl+Q to quit.
""".strip()


class HelpModal(ModalScreen):
    """Help screen showing commands and shortcuts."""

    def compose(self) -> ComposeResult:
        """Compose the help modal UI."""
        with Container(id="help-dialog"):
            yield Static("Paint the columns!", id="help-title")
            yield Static(HELP_TEXT, id="help-content")
            yield Button("Close (Esc)", id="close-help")

    @on(Button.Pressed, "#close-help")
    def close_help(self) -> None:
        """Close the help modal."""
        self.dismiss()

    def on_key(self, event: events.Key) -> None:
        """Handle key events for closing the help modal."""
        if event.key in ("escape", "question_mark", "f1"):
            self.dismiss()


class NoSamplesModal(ModalScreen):
    """Modal screen showing no samples are available."""

    def compose(self) -> ComposeResult:
        """Compose the no samples modal UI."""
        with Container(id="no-samples-dialog"):
            yield Static("No samples available", id="no-samples-title")
            yield Static(NO_SAMPLES_TEXT, id="no-samples-content")
            yield Button("Quit (Ctrl+Q)", id="quit-no-samples")

    @on(Button.Pressed, "#quit-no-samples")
    def quit_app(self) -> None:
        """Quit the application."""
        self.app.exit()

    def on_key(self, event: events.Key) -> None:
        """Handle key events for the no samples modal."""
        if event.key in ("escape", "ctrl+q", "ctrl+c"):
            self.app.exit()



"""Logging utilities."""

import functools
import importlib.metadata
import logging
import os
import time
from collections.abc import Callable
from typing import Any, Final, Literal, ParamSpec, TypeVar

import psutil
from rich.console import Console
from rich.progress import (
    BarColumn,
    Progress,
    SpinnerColumn,
    TextColumn,
    TimeElapsedColumn,
    TimeRemainingColumn,
)

_PLUGINS = None


def get_formatter() -> logging.Formatter:
    """Retrieve plugin registered in 'matchbox.logging' entry point, or fallback."""
    global _PLUGINS
    if _PLUGINS is None:
        _PLUGINS = []
        for ep in importlib.metadata.entry_points(group="matchbox.logging"):
            try:
                _PLUGINS.append(ep.load()())
            except Exception as e:  # noqa: BLE001
                logger.warning(f"Failed to load logging plugin: {e}")

    if _PLUGINS:
        return _PLUGINS[0]

    return logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")


LogLevelType = Literal["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]
"""Type for all Python log levels."""


class PrefixedLoggerAdapter(logging.LoggerAdapter):
    """A logger adapter that supports adding a prefix enclosed in square brackets.

    This adapter allows passing an optional prefix parameter to any logging call
    without modifying the underlying logger.
    """

    def process(
        self,
        msg: Any,  # noqa: ANN401
        kwargs: dict[str, Any],  # noqa: ANN401
    ) -> tuple[Any, dict[str, Any]]:  # noqa: ANN401
        """Process the log message, adding a prefix if provided.

        Args:
            msg: The log message
            kwargs: Additional arguments to the logging method

        Returns:
            Tuple of (modified_message, modified_kwargs)
        """
        prefix = kwargs.pop("prefix", None)

        if prefix:
            msg = f"[{prefix}] {msg}"

        return msg, kwargs


logger: Final[PrefixedLoggerAdapter] = PrefixedLoggerAdapter(
    logging.getLogger("matchbox"), {}
)
"""Logger for Matchbox.

Used for all logging in the Matchbox library.

Allows passing a prefix to any logging call.

Examples:
    ```python
    log_prefix = f"Model {name}"
    logger.debug("Inserting metadata", prefix=log_prefix)
    logger.debug("Inserting data", prefix=log_prefix)
    logger.info("Insert successful", prefix=log_prefix)
    ```
"""

console: Final[Console] = Console()
"""Console for Matchbox.

Used for any CLI utilities in the Matchbox library.
"""


def build_progress_bar(console_: Console | None = None) -> Progress:
    """Create a progress bar."""
    if console_ is None:
        console_ = console

    return Progress(
        SpinnerColumn(),
        TextColumn("[progress.description]{task.description}"),
        BarColumn(),
        TextColumn("[progress.percentage]{task.percentage:>3.0f}%"),
        TimeElapsedColumn(),
        TimeRemainingColumn(),
        console=console_,
    )


T = TypeVar("T")
P = ParamSpec("P")


def profile_time(
    logger: PrefixedLoggerAdapter = logger,
    level: int = logging.INFO,
    prefix: str | None = "Profiling",
    attr: str | None = None,
    kwarg: str | None = None,
) -> Callable[[Callable[P, T]], Callable[P, T]]:
    """Decorator to profile running time of functions and methods using logger.

    Args:
        logger: The logger to use.
        level: The level to use to log the profiling information. It defaults to INFO.
        prefix: Prefix to pass to the logged message.
        attr: Attribute name to extract from instantiated class.
        kwarg: Argument name to extract from function call.

    `attr` should be used when we want to include some class atribute in the log, e.g.
    node name in Source.
    `kwarg` should be used when we want to include the value passed to some function
    argument, e.g. path in `set_data`. This will only work if the argument is passed
    to the function as a kwarg, and not a positional argument.
    """

    def decorator(func: Callable[P, T]) -> Callable[P, T]:
        @functools.wraps(func)
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
            # If attr, will try to get its value from class instance
            if attr is not None:
                # If class, first argument will be self
                self = args[0] if args else None
                node = getattr(self, attr) if self and hasattr(self, attr) else None
            # If kwarg, will try to get value passed to function
            if kwarg is not None:
                value = kwargs.get(kwarg)

            start = time.perf_counter()
            try:
                return func(*args, **kwargs)
            finally:
                duration = time.perf_counter() - start

                if attr:
                    msg = f"`{func.__name__}` in node `{node}` took {duration:.3f}s"
                elif kwarg:
                    msg = (
                        f"`{func.__name__}` with {kwarg} `{value}` took {duration:.3f}s"
                    )
                else:
                    msg = f"`{func.__name__}` took {duration:.3f}s"

                logger.log(level, msg, prefix=prefix)

        return wrapper

    return decorator


def log_mem_usage(
    logger: PrefixedLoggerAdapter = logger,
    level: int = logging.INFO,
    prefix: str | None = "Profiling",
) -> None:
    """Log current memory usage for this process."""
    proc = psutil.Process(os.getpid())
    usage = proc.memory_info().rss / (1024**2)
    msg = f"Current memory used by process (MiB): {usage}"
    logger.log(level, msg, prefix=prefix)



"""Styling utilities for entity resolution evaluation UI."""

GROUP_STYLES = {
    # Row: q w e r t y u i o p
    "q": ("■", "#e53935"),  # red
    "w": ("●", "#43a047"),  # green
    "e": ("▲", "#1e88e5"),  # blue
    "r": ("◆", "#fdd835"),  # yellow
    "t": ("★", "#d81b60"),  # magenta
    "y": ("⬢", "#00acc1"),  # cyan
    "u": ("♦", "#fb8c00"),  # orange
    "i": ("▼", "#00897b"),  # teal
    "o": ("○", "#8e24aa"),  # purple
    "p": ("△", "#6d4c41"),  # brown
    # Row: a s d f g h j k l
    "a": ("◇", "#d81b60"),  # magenta
    "s": ("☆", "#00acc1"),  # cyan
    "d": ("⬡", "#e53935"),  # red
    "f": ("✦", "#43a047"),  # green
    "g": ("✧", "#1e88e5"),  # blue
    "h": ("⟐", "#fdd835"),  # yellow
    "j": ("✚", "#d81b60"),  # magenta
    "k": ("✖", "#3949ab"),  # indigo (keeps distance from teal above-left)
    "l": ("□", "#e53935"),  # red
    # Row: z x c v b n m
    "z": ("▽", "#fb8c00"),  # orange
    "x": ("◯", "#8e24aa"),  # purple (far from cyan/red above)
    "c": ("◻", "#fdd835"),  # yellow
    "v": ("◼", "#6d4c41"),  # brown
    "b": ("⬤", "#e53935"),  # red
    "n": ("☑", "#43a047"),  # green
    "m": ("✤", "#1e88e5"),  # blue
}


def get_group_style(group: str) -> tuple[str, str]:
    """Get symbol and colour for a group letter.

    Args:
        group: Single letter group identifier (a-z)

    Returns:
        Tuple of (symbol, colour)
    """
    return GROUP_STYLES.get(group.lower(), ("■", "white"))


def get_display_text(group: str, count: int) -> tuple[str, str]:
    """Get formatted display text with colour and symbol.

    Args:
        group: Single letter group identifier (a-z) or "unassigned"
        count: Number of items in this group

    Returns:
        Tuple of (formatted_text, colour)
    """
    if group == "unassigned":
        return f"UNASSIGNED ({count})", "dim"

    symbol, colour = get_group_style(group)
    text = f"{symbol} {group.upper()} ({count})"
    return text, colour


def generate_css_classes() -> str:
    """Generate CSS classes for all groups.

    Returns:
        CSS string with all group styling classes
    """
    lines = []
    for group, (_symbol, colour) in GROUP_STYLES.items():
        lines.append(f"ComparisonTable .group-{group} {{ background: {colour}; }}")
    return "\n".join(lines)



from io import BytesIO
from typing import TYPE_CHECKING, Any
from unittest.mock import Mock

import pyarrow as pa
import pyarrow.parquet as pq
import pytest
from botocore.exceptions import ClientError
from fastapi import UploadFile

from matchbox.common.arrow import table_to_buffer
from matchbox.common.dtos import ResolutionPath, ResolutionType
from matchbox.common.exceptions import (
    MatchboxServerFileError,
)
from matchbox.common.factories.models import model_factory
from matchbox.common.factories.sources import source_factory
from matchbox.server.uploads import (
    InMemoryUploadTracker,
    UploadTracker,
    file_to_s3,
    process_upload,
    s3_to_recordbatch,
)

if TYPE_CHECKING:
    from mypy_boto3_s3.client import S3Client
else:
    S3Client = Any


def test_file_to_s3(s3: S3Client) -> None:
    """Test that a file can be uploaded to S3."""
    # Create a mock bucket
    s3.create_bucket(
        Bucket="test-bucket",
        CreateBucketConfiguration={"LocationConstraint": "eu-west-2"},
    )
    source_testkit = source_factory(
        features=[
            {"name": "company_name", "base_generator": "company"},
            {"name": "address", "base_generator": "address"},
            {
                "name": "crn",
                "base_generator": "bothify",
                "parameters": (("text", "???-###-???-###"),),
            },
            {
                "name": "dh",
                "base_generator": "bothify",
                "parameters": (("text", "??######"),),
            },
        ],
    )
    all_companies = source_testkit.data.to_pandas()

    # Create a mock UploadFile
    all_companies["id"] = all_companies["id"].astype(str)
    table = pa.Table.from_pandas(all_companies)
    sink = pa.BufferOutputStream()
    pq.write_table(table, sink)
    file_content = sink.getvalue().to_pybytes()

    parquet_file = UploadFile(filename="test.parquet", file=BytesIO(file_content))

    # Call the function
    key = "foo.parquet"
    upload_id = file_to_s3(client=s3, bucket="test-bucket", key=key, file=parquet_file)
    # Validate response
    assert key == upload_id

    response_table = pa.Table.from_batches(
        [
            batch
            for batch in s3_to_recordbatch(
                client=s3, bucket="test-bucket", key=upload_id
            )
        ]
    )

    assert response_table.equals(table)


@pytest.fixture(scope="function")
def tracker_instance(request: pytest.FixtureRequest, tracker: str) -> UploadTracker:
    """Create a fresh tracker instance for each test."""
    tracker_obj = request.getfixturevalue(tracker)
    return tracker_obj


@pytest.mark.parametrize(
    "tracker",
    [
        pytest.param("upload_tracker_in_memory", id="in_memory"),
        pytest.param("upload_tracker_redis", id="redis"),
    ],
)
@pytest.mark.docker
class TestUploadTracker:
    @pytest.fixture(autouse=True)
    def setup(self, tracker_instance: str) -> None:
        self.tracker: UploadTracker = tracker_instance

    def test_getting_and_setting(self) -> None:
        assert self.tracker.get("upload_id") is None
        self.tracker.set("upload_id", "error_message")
        assert self.tracker.get("upload_id") == "error_message"


@pytest.mark.parametrize(
    "resolution_type",
    [
        pytest.param(ResolutionType.SOURCE, id="source"),
        pytest.param(ResolutionType.MODEL, id="model"),
    ],
)
def test_process_upload_success(s3: S3Client, resolution_type: ResolutionType) -> None:
    """Test that upload process hands data to backend."""
    # Prepare data
    if resolution_type == ResolutionType.SOURCE:
        testkit = source_factory().fake_run()
        resolution = testkit.source.to_resolution()
        data = testkit.data_hashes
    else:
        testkit = model_factory().fake_run()
        resolution = testkit.model.to_resolution()
        data = testkit.probabilities.to_arrow()

    # Setup mock backend and tracker
    mock_backend = Mock()
    mock_backend.settings.datastore.get_client.return_value = s3
    mock_backend.get_resolution = Mock(return_value=resolution)

    # Create bucket
    bucket = "test-bucket"
    test_key = "test-upload-id.parquet"

    s3.create_bucket(
        Bucket=bucket, CreateBucketConfiguration={"LocationConstraint": "eu-west-2"}
    )

    # Add parquet to S3 and verify
    buffer = table_to_buffer(data)
    s3.put_object(Bucket=bucket, Key=test_key, Body=buffer)

    assert s3.head_object(Bucket=bucket, Key=test_key)

    process_upload(
        backend=mock_backend,
        tracker=InMemoryUploadTracker(),
        s3_client=s3,
        upload_id="upload_id",
        bucket=bucket,
        filename=test_key,
        resolution_path=ResolutionPath(
            collection="collection", run=1, name="resolution"
        ),
    )

    # Verify file was deleted
    with pytest.raises(ClientError) as excinfo:
        s3.head_object(Bucket=bucket, Key=test_key)
    assert "404" in str(excinfo.value) or "NoSuchKey" in str(excinfo.value), (
        f"File was not deleted: {str(excinfo.value)}"
    )

    # Ensure data was inserted
    if resolution_type == ResolutionType.SOURCE:
        assert mock_backend.insert_source_data.called
    else:
        assert mock_backend.insert_model_data.called


def test_process_upload_empty_table(s3: S3Client) -> None:
    """Test that files representing empty table can be handled."""
    # Setup mock backend
    mock_backend = Mock()
    mock_backend.settings.datastore.get_client.return_value = s3

    # Create bucket
    bucket = "test-bucket"
    test_key = "test-upload-id.parquet"
    s3.create_bucket(
        Bucket=bucket,
        CreateBucketConfiguration={"LocationConstraint": "eu-west-2"},
    )

    # Add empty parquet to S3 and verify
    buffer = table_to_buffer(pa.Table.from_arrays([]))
    s3.put_object(Bucket=bucket, Key=test_key, Body=buffer)

    assert s3.head_object(Bucket=bucket, Key=test_key)

    # Trigger upload processing
    process_upload(
        backend=mock_backend,
        tracker=InMemoryUploadTracker(),
        s3_client=s3,
        upload_id="upload_id",
        bucket=bucket,
        filename=test_key,
        resolution_path=ResolutionPath(
            collection="collection", run=1, name="resolution"
        ),
    )

    # Verify file was deleted
    with pytest.raises(ClientError) as excinfo:
        s3.head_object(Bucket=bucket, Key=test_key)
    assert "404" in str(excinfo.value) or "NoSuchKey" in str(excinfo.value), (
        f"File was not deleted: {str(excinfo.value)}"
    )


def test_process_upload_deletes_file_on_failure(s3: S3Client) -> None:
    """Test that files are deleted from S3 even when processing fails.

    Other behaviours of this task are captured in the API integration tests for adding a
    source or a model.
    """
    # Prepare data
    source_testkit = source_factory().fake_run()

    # Setup mock backend and tracker
    tracker = InMemoryUploadTracker()
    mock_backend = Mock()
    mock_backend.settings.datastore.get_client.return_value = s3
    mock_backend.get_resolution = Mock(
        return_value=source_testkit.source.to_resolution()
    )
    mock_backend.insert_source_data = Mock(
        side_effect=ValueError("Simulated processing failure")
    )

    # Create bucket
    bucket = "test-bucket"
    test_key = "test-upload-id.parquet"

    s3.create_bucket(
        Bucket=bucket,
        CreateBucketConfiguration={"LocationConstraint": "eu-west-2"},
    )

    # Add parquet to S3 and verify
    buffer = table_to_buffer(source_testkit.data_hashes)
    s3.put_object(Bucket=bucket, Key=test_key, Body=buffer)

    assert s3.head_object(Bucket=bucket, Key=test_key)

    # Run the process, expecting it to fail
    with pytest.raises(MatchboxServerFileError) as excinfo:
        process_upload(
            backend=mock_backend,
            tracker=tracker,
            s3_client=s3,
            upload_id="upload_id",
            bucket=bucket,
            filename=test_key,
            resolution_path=ResolutionPath(
                collection="collection", run=1, name="resolution"
            ),
        )

    assert "Simulated processing failure" in str(excinfo.value)

    # Check that the error was added to tracker
    error = tracker.get("upload_id")
    assert "Simulated processing failure" in error

    # Verify file was deleted despite the failure
    with pytest.raises(ClientError) as excinfo:
        s3.head_object(Bucket=bucket, Key=test_key)
    assert "404" in str(excinfo.value) or "NoSuchKey" in str(excinfo.value), (
        f"File was not deleted: {str(excinfo.value)}"
    )

    # Ensure resolution is marked as ready again
    assert mock_backend.unlock_resolution_data.called
